/*
 * (c) Copyright 2002 - 2004 -- Anders Torger
 *
 * This program is open source. For license terms, see the LICENSE file.
 *
 */
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <dirent.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <signal.h>
#include <fcntl.h>
#include <math.h>

#include "defs.h"
#include "elog.h"
#include "sysmon.h"
#include "bfir.h"
#include "bit.h"
#include "bfirmem.h"
#include "idarray.h"
#include "almusvcu.h"
#include "bfirdata.h"
#include "bfirconf.h"

#include <bfmod.h>

/*
 * FIXME: move config generation to another file?
 * FIXME: can BF_MAXFILTERS be exceeded?
 */

#warning cleanup input channel count translation
#warning check for hard settings and soft settings reverb channel mismatch
#warning better eq benchmarking (handle modes)

static struct {
    char path[PATH_MAX];
    char modules_path[PATH_MAX];
    char convolver_config[PATH_MAX];
    bool_t apply_dither;
    bool_t allow_poll_mode;
    int channel_count[2][VCU_CH_COUNT];
    int ambisonic_input_format;

    struct  {
        double ampl[2][BFIR_MAX_OUTPUT_CHANNELS];
        bool_t muted[2][BFIR_MAX_OUTPUT_CHANNELS];
        int delays[2][BFIR_MAX_OUTPUT_CHANNELS];
    } virt;

    struct {
        double ampl[2][VCU_CH_WP_COUNT][BFIR_MAX_OUTPUT_CHANNELS];
        bool_t muted[2][VCU_CH_WP_COUNT][BFIR_MAX_OUTPUT_CHANNELS];
        int delays[2][VCU_CH_WP_COUNT][BFIR_MAX_OUTPUT_CHANNELS];
    } actu;
    
    struct runtime_map runtime_map;
    struct chmap chmap;
    int device_count[2];
    char device[2][CONF_DEVICES_MAX][PATH_MAX];
    char device_params[2][CONF_DEVICES_MAX][PATH_MAX];
    int sample_format[2];
    int open_channels[2][CONF_DEVICES_MAX];
    int tot_open_channels[2];

    uint32_t eq_active[2][VCU_CH_WP_COUNT][VCU_MAX_CHANNELS/32];
    int eq_id[2][VCU_CH_WP_COUNT][VCU_MAX_CHANNELS];
    bool_t xtalk_cancel[BFIR_MAX_OUTPUT_CHANNELS/2];
    int xtalkcancel_id[BFIR_MAX_OUTPUT_CHANNELS/2];
    int reverb_id;
    double ambisonic_gains[9 * BFIR_MAX_OUTPUT_CHANNELS];
    int dyneq_count;
    struct bfirdyneq *dyneq[VCU_CH_COUNT * BFIR_MAX_OUTPUT_CHANNELS];
} brutefir_settings;

static struct {
    int realsize;
    int n_rates;
    int rates[FILTERPRG_RATES_MAX];
    int block_sizes[FILTERPRG_RATES_MAX];
    /* some hard settings is in brutefir_settings */
} hard_settings;

bool_t bfir_debug = false;

static char log_path[PATH_MAX];
static int ext_channel_order[2][CONF_DEVICES_MAX][VCU_MAX_CHANNELS];
static int hardware_min_blocksize;
static int filterprgs_count[BFIR_FILTER_TYPE_COUNT];
static FILE *brutefir_instream = NULL;
static FILE *brutefir_outstream = NULL;
static pid_t brutefir_pid = 0;
static void (*bfir_peak_meter)(double peak);
static void (*bfir_process_terminated)(int reason);
static bool_t hard_settings_set = false;
static bool_t soft_settings_set = false;
static bool_t alloc_benchrun_filter_failed;
static int longest_benchrun_filter_blocks;

static int buffer_commands = 0;
static char *cmdbuffer = NULL;
static int cmdbuffer_cap = 0;
static int cmdbuffer_pos = 0;

#define MUST_NOT_BE_RUNNING(retval)                                            \
    if (brutefir_instream != NULL) {                                           \
        elog(LOGDEBUG, "cannot do this while BruteFIR is running");            \
        return retval;                                                         \
    }

#define MUST_BE_RUNNING(retval)                                                \
    if (brutefir_instream == NULL) {                                           \
        elog(LOGDEBUG, "cannot do this while BruteFIR is not running");        \
        return retval;                                                         \
    }

static int
translate_exitcode(int exitcode)
{
    elog(LOGDEBUG, "exitcode: %d", exitcode);
    
    switch (exitcode) {
    case BF_EXIT_OK:
	return 0;
    case BF_EXIT_OTHER:
	return VCU_ERR_CONVOLVER_FAILED_UNKNOWN;
    case BF_EXIT_INVALID_CONFIG:
	return VCU_ERR_CONVOLVER_FAILED_UNKNOWN;
    case BF_EXIT_NO_MEMORY:
	return VCU_ERR_CONVOLVER_FAILED_NO_MEMORY;
    case BF_EXIT_INVALID_INPUT:
	return VCU_ERR_CONVOLVER_FAILED_SIGNAL;
    case BF_EXIT_BUFFER_UNDERFLOW:
	return VCU_ERR_CONVOLVER_FAILED_UNDERFLOW;
    default:
	return VCU_ERR_CONVOLVER_FAILED_UNKNOWN;
    }
}

static bool_t
isruntime(void)
{
    return brutefir_instream != NULL;
}

static void
putcommand(const char cmd[])
{
    char path[PATH_MAX + 100];
    FILE *stream;
    int slen;
    
    if (!isruntime() || cmd[0] == '\0') {
	return;
    }
    if (buffer_commands > 0) {
        slen = strlen(cmd);
        if (cmdbuffer_cap < cmdbuffer_pos + slen + 1) {
            cmdbuffer_cap = cmdbuffer_pos + slen + 1;
            cmdbuffer = realloc(cmdbuffer, cmdbuffer_cap);
            if (cmdbuffer == NULL) {
                elog(LOGERROR, "memory allocation failure");
                exit(EXIT_FAILURE);
            }
        }
        memcpy(&cmdbuffer[cmdbuffer_pos], cmd, slen + 1);
        cmdbuffer_pos += slen;
        return;
    }
    elog(LOGDEBUG, "putting command: (%s)", cmd);
    if (bfir_debug) {
        sprintf(path, "%s%s", log_path, BFIR_DEBUG_BRUTEFIR_CLI_INPUT);
        if ((stream = fopen(path, "at")) != NULL) {
	    fprintf(stream, "%s\n", cmd);
	    fclose(stream);
	}
    }
    fprintf(brutefir_instream, "%s\n", cmd);
}

static int
build_maps(const struct bfir_chmap *map[2],
           int ambisonic_input_format,
           struct chmap *chmap,
           int channel_count[2][VCU_CH_COUNT])
{
    uint32_t logimask[VCU_CH_COUNT][VCU_MAX_CHANNELS/32];
    int io, phys, type, count, virt, n, logi;

    memset(channel_count, 0, sizeof(int) * VCU_CH_COUNT);
    memset(chmap, 0, sizeof(struct chmap));
    memset(chmap->virt2phys, -1, sizeof(chmap->virt2phys));
    memset(chmap->virt2logi_type, -1, sizeof(chmap->virt2logi_type));
    memset(chmap->virt2logi_indx, -1, sizeof(chmap->virt2logi_indx));
    FOR_IN_AND_OUT {
        memset(logimask, 0, sizeof(logimask));
        count = 0;
        virt = 0;
        for (phys = 0; phys < brutefir_settings.tot_open_channels[io]; phys++) {
            for (type = 0; type < VCU_CH_COUNT; type++) {
                logi = -1;
                while ((logi = bit_find(map[io][phys].channels[type], logi + 1,
                                        VCU_MAX_CHANNELS - 1)) != -1)
                {
                    if (virt >= VCU_MAX_CHANNELS) {
                        elog(LOGDEBUG, "Too many channels");
                        return VCU_ERR_CONVOLVER_FAILED_TOO_COMPLEX;
                    }
                    bit_set(chmap->phys2virt[io][phys], virt);
                    elog(LOGDEBUG, "virt2phys[%d][%d] = %d", io, virt, phys);
                    chmap->virt2phys[io][virt] = phys;
                    chmap->virt2logi_type[io][virt] = type;
                    chmap->virt2logi_indx[io][virt] = logi;
                    elog(LOGDEBUG, "logi2virt[%d][%d][%d] = %d",
                         io, type, logi, virt);
                    bit_set(chmap->logi2virt[io][type][logi], virt);
                    bit_set(logimask[type], logi);
                    virt++;
                }
            }
        }
        chmap->virt_count[io] = virt;

        for (type = 0; type < VCU_CH_COUNT; type++) {
            n = -1;
            count = 0;
            while ((n = bit_find(logimask[type], n + 1,
                                 VCU_MAX_CHANNELS - 1)) != -1)
            {
                if (n != count) {
                    elog(LOGDEBUG, "invalid channel mapping for type %d", type);
                    return VCU_ERR_INVALID_VALUE;
                }
                count++;
            }
            channel_count[io][type] = count;
        }
    }

    switch (ambisonic_input_format) {
    case VCU_AMBISONIC_INPUT_FORMAT_FIRST_ORDER:
        count = 4;
        break;
    case VCU_AMBISONIC_INPUT_FORMAT_SECOND_ORDER:
        count = 9;
        break;
    case VCU_AMBISONIC_INPUT_FORMAT_UHJ:
        count = 2;
        break;
    default:
        count = 0;
        break;
    }

    if ((channel_count[OUT][VCU_CH_REVERB] > 0 &&
         channel_count[IN][VCU_CH_REVERB] != 1 &&
         channel_count[IN][VCU_CH_REVERB] != 2 &&
         channel_count[IN][VCU_CH_REVERB] != 3) ||
        (channel_count[OUT][VCU_CH_AMBIOPOLE] !=
         channel_count[IN][VCU_CH_AMBIOPOLE]) ||
        (channel_count[OUT][VCU_CH_PASSTHRU] !=
         channel_count[IN][VCU_CH_PASSTHRU]) ||
        (channel_count[OUT][VCU_CH_AMBISONIC] > 0 &&
         channel_count[IN][VCU_CH_AMBISONIC] != count))
    {
        FOR_IN_AND_OUT {
            for (type = 0; type < VCU_CH_COUNT; type++) {
                elog(LOGDEBUG, "channel count io %d, type %d: %d",
                     io, type, channel_count[io][type]);
            }
        }
        elog(LOGDEBUG, "invalid channel mapping");
        return VCU_ERR_INVALID_VALUE;
    }

    return 0;
}

static void
apply_mdv(int io,
          int virt,
          int phys,
          int type,
          int logi)
{
    struct runtime_map *rmap;
    double ampl, scale;
    int delay, filt;
    uint32_t *fmap;
    char str[100];
    bool_t mute;

    rmap = &brutefir_settings.runtime_map;
    delay = brutefir_settings.actu.delays[io][type][logi] +
        brutefir_settings.actu.delays[io][VCU_CH_PHYSICAL][phys];
    mute = brutefir_settings.actu.muted[io][type][logi] |
        brutefir_settings.actu.muted[io][VCU_CH_PHYSICAL][phys];
    ampl = brutefir_settings.actu.ampl[io][type][logi] +
        brutefir_settings.actu.ampl[io][VCU_CH_PHYSICAL][phys];
    scale = 1.0;

    if (delay != brutefir_settings.virt.delays[io][virt]) {
        brutefir_settings.virt.delays[io][virt] = delay;
        if (isruntime()) {
            sprintf(str, "c%cd %d %d;", io == IN ? 'i' : 'o', virt, delay);
            putcommand(str);
        }
    }
    if (!mute != !brutefir_settings.virt.muted[io][virt]) {
        brutefir_settings.virt.muted[io][virt] = !!mute;
        if (isruntime()) {
            sprintf(str, "tm%c %d;", io == IN ? 'i' : 'o', virt);
            putcommand(str);
        }
    }
    if (ampl != brutefir_settings.virt.ampl[io][virt]) {
        brutefir_settings.virt.ampl[io][virt] = ampl;
        if (isruntime()) {
            filt = -1;
            fmap = rmap->virt2filt[io][virt];
            while ((filt = bit_find(fmap, filt + 1, BF_MAXFILTERS - 1)) != -1) {
                if (io == OUT) {
                    scale = rmap->out_scales[filt][virt];
                } else {
                    scale = 1.0;
                }
                
                if (scale != 1.0) {
                    scale *= pow(10, ampl / 20);
                    sprintf(str, "cf%ca %d %d m%g;", io == IN ? 'i' : 'o',
                            filt, virt, scale);
                } else {
                    sprintf(str, "cf%ca %d %d %g;", io == IN ? 'i' : 'o',
                            filt, virt, -ampl);
                }
                
                putcommand(str);
            }
        }
    }
}

static void
apply_actu2virt_mdv(int io,
                    int type,
                    const uint32_t mask[],
                    int start,
                    int end)
{
    int virt, phys, logi, count;
    struct chmap *chmap;
    uint32_t *vmap;

    chmap = &brutefir_settings.chmap;
    if (type == VCU_CH_PHYSICAL) {
        bfir_start_buffering_commands();
        count = brutefir_settings.tot_open_channels[io];
        phys = start - 1;
        while ((phys = bit_find(mask, phys + 1, end)) != -1) {
            vmap = chmap->phys2virt[io][phys];
            virt = -1;
            while ((virt = bit_find(vmap, virt + 1,
                                    VCU_MAX_CHANNELS - 1)) != -1)
            {
                type = chmap->virt2logi_type[io][virt];
                logi = chmap->virt2logi_indx[io][virt];
                apply_mdv(io, virt, phys, type, logi);
            }            
        }
        bfir_commit_buffered_commands();    
        return;
    }
    
    bfir_start_buffering_commands();
    count = brutefir_settings.channel_count[io][type];
    logi = start - 1;
    while ((logi = bit_find(mask, logi + 1, end)) != -1) {
        vmap = brutefir_settings.chmap.logi2virt[io][type][logi];
        virt = -1;
        while ((virt = bit_find(vmap, virt + 1, VCU_MAX_CHANNELS - 1)) != -1) {
            phys = chmap->virt2phys[io][virt];
            apply_mdv(io, virt, phys, type, logi);
        }
    }
    bfir_commit_buffered_commands();
}

static int
alloc_benchrun_filter(int n_blocks,
                      int block_size)
{
    int id;

    if (n_blocks > longest_benchrun_filter_blocks) {
        longest_benchrun_filter_blocks = n_blocks;
    }
    id = bfirdata_new_dummy_filter(n_blocks, block_size);
    if (id == -1) {
        alloc_benchrun_filter_failed = true;
    }
    return id;
}

static bool_t
benchrun(int sample_rate,
         bool_t apply_dither,
	 int block_size,	 
	 int n_reverb_blocks,         
	 const int n_xtalk_cancel_blocks[],
	 const int *n_equalisation_blocks[2][VCU_CH_WP_COUNT],
         int ambisonic_input_format,
         const struct bfir_chmap *maps[2],
	 int *brutefir_exit_status,
	 double *rti)
{
    int fd, n, i, io, in_pipe[2], out_pipe[2], rti_counter, type, count;
    int newmap[VCU_MAX_CHANNELS], n_coeffs;
    char str[1000], path[PATH_MAX + 100];
    const struct brutefir_coeff *coeffs;
    struct brutefir_config config;
    FILE *stream, *debug_stream;
    pid_t pid;

    alloc_benchrun_filter_failed = false;
    longest_benchrun_filter_blocks = 1;
    bfirdata_reset();
    
    memset(&config, 0, sizeof(config));

    config.sample_rate = sample_rate;
    config.block_length = block_size;
    config.n_processors = sysmon_n_cpus();
    config.show_progress = true;
    config.apply_dither = apply_dither;
    config.powersave = false;
    config.watch_clock = false;
    config.double_precision = !!(hard_settings.realsize > 4);
    config.cli_pipe[0] = -1;
    config.cli_pipe[1] = -1;
    strcpy(config.modules_path, brutefir_settings.modules_path);
    strcpy(config.convolver_config, brutefir_settings.convolver_config);
    config.ambisonic_input_format = ambisonic_input_format;
    config.max_delay[IN] = 0;
    config.max_delay[OUT] = 0;
    
    if (build_maps(maps, ambisonic_input_format,
                   &config.chmap, config.channel_count) != 0)
    {
        return false;
    }

    config.device_count[IN] = 1;
    config.device_count[OUT] = 1;
    strcpy(config.device[IN][0], BFIR_BENCHMARK_INPUT_DEVICE);
    strcpy(config.device[OUT][0], BFIR_BENCHMARK_OUTPUT_DEVICE);
    strcpy(config.device_params[IN][0], BFIR_BENCHMARK_INPUT_DEVICE_PARAMS);
    strcpy(config.device_params[OUT][0], BFIR_BENCHMARK_OUTPUT_DEVICE_PARAMS);
    FOR_IN_AND_OUT {

        for (n = 0; n < VCU_MAX_CHANNELS; n++) {
            config.ext_channel_order[io][0][n] = n;
        }
        
        /* remap physical channels */
        for (n = i = 0; n < brutefir_settings.tot_open_channels[io]; n++) {
            if (bit_find(config.chmap.phys2virt[io][n], 0,
                         config.chmap.virt_count[io] - 1) != -1)
            {
                newmap[n] = i;
                i++;
            }                         
        }
        config.open_channels[io][0] = i;
        config.tot_open_channels[io] = i;

        memset(config.chmap.phys2virt[io], 0,
               sizeof(config.chmap.phys2virt[io]));
        for (n = 0; n < config.chmap.virt_count[io]; n++) {
            config.chmap.virt2phys[io][n] =
                newmap[config.chmap.virt2phys[io][n]];
            bit_set(config.chmap.phys2virt[io][config.chmap.virt2phys[io][n]],
                    n);
        }
        
        config.sample_format[io] = brutefir_settings.sample_format[io];
    }

    elog(LOGDEBUG, "equaliser filters");
    FOR_IN_AND_OUT {
        for (type = 0; type < VCU_CH_WP_COUNT; type++) {
            if (type != VCU_CH_PHYSICAL) {
                count = config.channel_count[io][type];
            } else {
                count = config.tot_open_channels[io];
            }
            for (n = 0; n < count; n++) {
                if (n_equalisation_blocks[io][type][n] > 0) {
                    bit_set(config.eqactive[io][type], n);
                    config.eqcoeffs[io][type][n] = alloc_benchrun_filter
                        (n_equalisation_blocks[io][type][n], block_size);
                } else {
                    config.eqcoeffs[io][type][n]= -1;
                }
            }
        }
    }
    
    /* ambiopole stuff */
    elog(LOGDEBUG, "ambiopole filters");
    for (n = 0; n < config.channel_count[OUT][VCU_CH_AMBIOPOLE] / 2; n++) {
        config.xtalk_cancel[n] = (n_xtalk_cancel_blocks[n] > 0);
        for (i = 0; i < 4; i++) {
            if (config.xtalk_cancel[n]) {
                config.xtalkcancel_coeffs[4*n+i] =
                    alloc_benchrun_filter(n_xtalk_cancel_blocks[n], block_size);
            } else {
                config.xtalkcancel_coeffs[4*n+i] = -1;
            }
        }
    }
    /* reverb stuff */
    elog(LOGDEBUG, "reverb filters");
    for (n = 0; n < config.channel_count[OUT][VCU_CH_REVERB]; n++) {
        for (i = 0; i < config.channel_count[IN][VCU_CH_REVERB]; i++) {
            config.reverb_coeffs[config.channel_count[IN][VCU_CH_REVERB]*n+i] =
                alloc_benchrun_filter(n_reverb_blocks, block_size);
        }
    }
    /* ambisonic stuff */
    if (config.channel_count[OUT][VCU_CH_AMBISONIC] > 0) {
        memcpy(config.ambisonic_gains, brutefir_settings.ambisonic_gains,
               sizeof(config.ambisonic_gains));
    }
    config.n_blocks = longest_benchrun_filter_blocks;
        
    *rti = 0.0;
    if (alloc_benchrun_filter_failed) {
	elog(LOGDEBUG, "Failed to allocate benchrun filters.");	
        bfirdata_reset();
        *brutefir_exit_status = BF_EXIT_NO_MEMORY;
        return true;
    }
    
    elog(LOGDEBUG, "getting dummy coefficients");
    coeffs = bfirdata_get_dummy_coeffs(&n_coeffs);
    elog(LOGDEBUG, "got %d dummy filters", n_coeffs);
    
    if (pipe(in_pipe) == -1 || pipe(out_pipe) == -1) {
	elog(LOGDEBUG, "Could not create pipe: %s.", strerror(errno));	
        bfirdata_reset();
	return false;
    }
    fd = fcntl(in_pipe[0], F_DUPFD, 3);
    close(in_pipe[0]);
    in_pipe[0] = fd;
    fd = fcntl(out_pipe[1], F_DUPFD, 3);
    close(out_pipe[1]);
    out_pipe[1] = fd;

    elog(LOGDEBUG, "forking");
    switch ((pid = fork())) {
    case 0:
	close(in_pipe[1]);	
	dup2(in_pipe[0], STDIN_FILENO);
	dup2(out_pipe[1], STDOUT_FILENO);
	dup2(out_pipe[1], STDERR_FILENO);
        bfirmem_finish();
	execl(brutefir_settings.path, brutefir_settings.path, "-nodefault",
              "stdin", NULL);
	elog(LOGERROR, "Exec failed: %s.", strerror(errno));
	exit(BF_EXIT_OTHER);
	break;
    case -1:
	elog(LOGERROR, "Fork failed: %s.", strerror(errno));
        bfirdata_reset();
	return false;
    default:
	break;
    }        
    close(in_pipe[0]);
    close(out_pipe[1]);

    stream = fdopen(in_pipe[1], "w");
    elog(LOGDEBUG, "writing config");
    if (!bfirconf_config_write(stream, &config, coeffs, n_coeffs,
                               &brutefir_settings.runtime_map))
    {
	fclose(stream);
	close(out_pipe[0]);
	kill(pid, SIGTERM);
	wait(&n);
        bfirdata_reset();
	return false;
    }	
    fclose(stream);
    elog(LOGDEBUG, "done writing config");
    if (bfir_debug) {
        sprintf(path, "%s%s", log_path, BFIR_DEBUG_LAST_BRUTEFIR_CONFIG);
	if ((stream = fopen(path, "wt")) != NULL) {
	    bfirconf_config_write(stream, &config, coeffs, n_coeffs,
                                  &brutefir_settings.runtime_map);
	    fclose(stream);
            elog_concatenate_file(path);
	}
    }
    
    stream = fdopen(out_pipe[0], "r");
    sleep(3);
    rti_counter = 0;
    if (bfir_debug) {
        sprintf(path, "%s%s", log_path, BFIR_DEBUG_BRUTEFIR_CONSOLE_OUTPUT);
	debug_stream = fopen(path, "wt");
    } else {
	debug_stream = NULL;
    }
    elog(LOGDEBUG, "console data");
    str[999] = '\0';
    while (fgets(str, 999, stream) != NULL) {
	if (debug_stream != NULL) {
	    fprintf(debug_stream, "%s", str);
	}
	if (strstr(str, "rti: ") == str) {
            elog(LOGDEBUG, "got rti line");
	    if (++rti_counter == BFIR_BENCHMARK_RTI_COUNTER) {                
		*rti = atof(&str[4]);
		kill(pid, SIGTERM);
	    }
	}
    }
    if (debug_stream != NULL) {
	fclose(debug_stream);
        elog_concatenate_file(path);
    }
    fclose(stream);
    wait(brutefir_exit_status);

    elog(LOGDEBUG, "rti was %f", *rti);

    *brutefir_exit_status = WEXITSTATUS(*brutefir_exit_status);
    
    bfirdata_reset();
    if (*brutefir_exit_status == BF_EXIT_OK && *rti == 0.0) {
        return false;
    }
    return true;
}

bool_t
bfir_init(int alloc_bytes,
          const char _log_path[],
	  const char brutefir_path[],
          const char brutefir_modules_path[],
          const char brutefir_convolver_config[],
          int brutefir_input_device_count,
	  char brutefir_input_device[CONF_DEVICES_MAX][PATH_MAX],
	  char brutefir_input_device_params[CONF_DEVICES_MAX][PATH_MAX],
	  const int brutefir_input_open_channels[CONF_DEVICES_MAX],
	  int brutefir_input_channel_map[CONF_DEVICES_MAX][VCU_MAX_CHANNELS],
          int brutefir_output_device_count,
	  char brutefir_output_device[CONF_DEVICES_MAX][PATH_MAX],
	  char brutefir_output_device_params[CONF_DEVICES_MAX][PATH_MAX],
	  const int brutefir_output_open_channels[CONF_DEVICES_MAX],
	  int brutefir_output_channel_map[CONF_DEVICES_MAX][VCU_MAX_CHANNELS],
          int min_blocksize,
          bool_t allow_poll_mode)
{
    static bool_t init = false;
    int n;

    /* Ignore SIGPIPE, we can receive this if BruteFIR dies prematurely */
    signal(SIGPIPE, SIG_IGN);

    memset(&brutefir_settings, 0, sizeof(brutefir_settings));
    memset(brutefir_settings.eq_id, 0xFF,
           sizeof(brutefir_settings.eq_id));
    memset(filterprgs_count, 0, sizeof(filterprgs_count));

    hardware_min_blocksize = min_blocksize;
    
    if (!init) {
	init = true;
    } else {
	elog(LOGERROR, "Already initialised.");
	return false;
    }
    brutefir_settings.device_count[IN] = brutefir_input_device_count;
    brutefir_settings.tot_open_channels[IN] = 0;    
    for (n = 0; n < brutefir_settings.device_count[IN]; n++) {
        brutefir_settings.open_channels[IN][n] =
            brutefir_input_open_channels[n];
        brutefir_settings.tot_open_channels[IN] +=
            brutefir_input_open_channels[n];
        strncpy(brutefir_settings.device[IN][n], brutefir_input_device[n],
                PATH_MAX);
        brutefir_settings.device[IN][n][PATH_MAX - 1] = '\0';
        strncpy(brutefir_settings.device_params[IN][n],
                brutefir_input_device_params[n], PATH_MAX);
        brutefir_settings.device_params[IN][n][PATH_MAX - 1] = '\0';
        memcpy(ext_channel_order[IN][n], brutefir_input_channel_map[n],
               VCU_MAX_CHANNELS * sizeof(int));
    }
    
    brutefir_settings.device_count[OUT] = brutefir_output_device_count;
    brutefir_settings.tot_open_channels[OUT] = 0;
    for (n = 0; n < brutefir_settings.device_count[OUT]; n++) {
        brutefir_settings.open_channels[OUT][n] =
            brutefir_output_open_channels[n];
        brutefir_settings.tot_open_channels[OUT] +=
            brutefir_output_open_channels[n];
        strncpy(brutefir_settings.device[OUT][n], brutefir_output_device[n],
                PATH_MAX);
        brutefir_settings.device[OUT][n][PATH_MAX - 1] = '\0';
        strncpy(brutefir_settings.device_params[OUT][n],
                brutefir_output_device_params[n], PATH_MAX);
        brutefir_settings.device_params[OUT][n][PATH_MAX - 1] = '\0';
        memcpy(ext_channel_order[OUT][n], brutefir_output_channel_map[n],
               VCU_MAX_CHANNELS * sizeof(int));
    }
    /* FIXME: add some more sanity checks */
    if (brutefir_settings.tot_open_channels[IN] < 2) {
	elog(LOGERROR, "Too few input channels (%d)",
		brutefir_settings.tot_open_channels[IN]);
        return false;
    }

    strncpy(log_path, _log_path, PATH_MAX);
    log_path[PATH_MAX - 1] = '\0';
    strncpy(brutefir_settings.path, brutefir_path, PATH_MAX);
    brutefir_settings.path[PATH_MAX - 1] = '\0';
    strncpy(brutefir_settings.modules_path, brutefir_modules_path, PATH_MAX);
    brutefir_settings.modules_path[PATH_MAX - 1] = '\0';
    strncpy(brutefir_settings.convolver_config, brutefir_convolver_config,
            PATH_MAX);
    brutefir_settings.convolver_config[PATH_MAX - 1] = '\0';
    brutefir_settings.allow_poll_mode = !!allow_poll_mode;

    return bfirdata_init(alloc_bytes);
}

void
bfir_flush_logs(void)
{
    char path[PATH_MAX + 100];
    
    sprintf(path, "%s%s", log_path, BFIR_DEBUG_BRUTEFIR_CONSOLE_OUTPUT);
    remove(path);
    sprintf(path, "%s%s", log_path, BFIR_DEBUG_LAST_BRUTEFIR_CONFIG);
    remove(path);
    sprintf(path, "%s%s", log_path, BFIR_DEBUG_BRUTEFIR_CLI_OUTPUT);
    remove(path);
    sprintf(path, "%s%s", log_path, BFIR_DEBUG_BRUTEFIR_CLI_INPUT);
    remove(path);
}

bool_t
bfir_hard_settings(int n_rates,
		   const int rates[],
		   const int block_sizes[],
		   int n_reverb_input_channels,
		   int n_reverb_output_channels,
                   int ambisonic_input_format,
                   int realsize)
{
    int n;

    /* apply settings */
    MUST_NOT_BE_RUNNING(false);
    if (n_rates < 1 || n_rates > FILTERPRG_RATES_MAX) {
	elog(LOGDEBUG, "invalid number of rates: %d\n", n_rates);
	return false;
    }
    if (n_reverb_output_channels < 0 ||
        n_reverb_output_channels > VCU_MAX_CHANNELS)
    {
	elog(LOGDEBUG, "invalid number of reverb output channels: %d\n",
		n_reverb_output_channels);
	return false;
    }
    if (n_reverb_input_channels < 0 ||
        n_reverb_input_channels > 3 ||
        (n_reverb_output_channels > 0 && n_reverb_input_channels == 0))
    {
	elog(LOGDEBUG, "invalid number of reverb input channels: %d\n",
		n_reverb_input_channels);
	return false;
    }
    hard_settings.realsize = realsize;
    hard_settings.n_rates = n_rates;
    brutefir_settings.channel_count[IN][VCU_CH_REVERB] =
        n_reverb_input_channels;
    brutefir_settings.channel_count[OUT][VCU_CH_REVERB] =
        n_reverb_output_channels;
    brutefir_settings.ambisonic_input_format = ambisonic_input_format;
    for (n = 0; n < n_rates; n++) {
	hard_settings.rates[n] = rates[n];
	hard_settings.block_sizes[n] = block_sizes[n];
    }

    bfirdata_settings(n_reverb_input_channels,
                      n_reverb_output_channels,
                      n_rates,
                      rates,
                      block_sizes,
                      realsize);

    hard_settings_set = true;
    return true;
}

int
bfir_verify_phys_map(const struct bfir_chmap *maps[2],
                     int ambisonic_input_format,
                     int channel_count[2][VCU_CH_COUNT])
{
    struct chmap chmap;

    if (maps == NULL || maps[IN] == NULL || maps[OUT] == NULL) {
        return VCU_ERR_INVALID_VALUE;
    }
    return build_maps(maps, ambisonic_input_format, &chmap, channel_count);
}

bool_t
bfir_soft_settings(int input_sample_format,
		   int output_sample_format,
                   bool_t apply_dither,
		   uint32_t apply_eq[2][VCU_CH_WP_COUNT][VCU_MAX_CHANNELS/32],
		   const bool_t apply_xtc[],
                   const double ambisonic_gains[],
		   const struct bfir_chmap *maps[2])
{
    int channel_count[2][VCU_CH_COUNT];
    struct chmap chmap;

    MUST_NOT_BE_RUNNING(false);
    if (!hard_settings_set) {
	elog(LOGERROR, "hard settings must be set");
	return false;
    }

    if (build_maps(maps, brutefir_settings.ambisonic_input_format,
                   &chmap, channel_count) != 0)
    {
        return false;
    }
    brutefir_settings.chmap = chmap;
    memcpy(brutefir_settings.channel_count, channel_count,
           sizeof(channel_count));

    memcpy(brutefir_settings.eq_active, apply_eq,
           2 * VCU_CH_WP_COUNT * VCU_MAX_CHANNELS / 32 * sizeof(uint32_t));
    
    brutefir_settings.sample_format[IN] = input_sample_format;
    brutefir_settings.sample_format[OUT] = output_sample_format;
    brutefir_settings.apply_dither = apply_dither;
    if (brutefir_settings.channel_count[OUT][VCU_CH_AMBISONIC] > 0) {
        memcpy(brutefir_settings.ambisonic_gains,
               ambisonic_gains,
               brutefir_settings.channel_count[IN][VCU_CH_AMBISONIC] *
               brutefir_settings.channel_count[OUT][VCU_CH_AMBISONIC] *
               sizeof(double));
    }
    
    memset(brutefir_settings.xtalk_cancel, 0,
           sizeof(brutefir_settings.xtalk_cancel));
    if (channel_count[OUT][VCU_CH_AMBIOPOLE] > 0) {
        memcpy(brutefir_settings.xtalk_cancel, apply_xtc,
               channel_count[OUT][VCU_CH_AMBIOPOLE] / 2 * sizeof(bool_t));
    }
    
    memset(&brutefir_settings.virt, 0, sizeof(brutefir_settings.virt));
    memset(&brutefir_settings.actu, 0, sizeof(brutefir_settings.actu));

    soft_settings_set = true;
    return true;
}

int
bfir_new_filters(int type,
                 struct bfirfilter filters[],
                 const int lengths[])
{
    MUST_NOT_BE_RUNNING(-1);

    return bfirdata_new_filters(type, filters, lengths);
}

int
bfir_shorten_filters(int id,
                     const int lengths[])
{
    MUST_NOT_BE_RUNNING(VCU_ERR_PREREQUISITES_NOT_FULFILLED);

    return bfirdata_shorten_filters(id, lengths);
}

void
bfir_delete_filters(int id)
{
    MUST_NOT_BE_RUNNING( );

    bfirdata_delete_filters(id);
}

double
bfir_filter_capacity(int type,
                     double *max_capacity)
{
    return bfirdata_filter_capacity(type, max_capacity);
}

void
bfir_set_reverbprg(int id)
{
    char cmd[brutefir_settings.channel_count[IN][VCU_CH_REVERB] *
            brutefir_settings.channel_count[OUT][VCU_CH_REVERB] * 20];
    char *str;
    int n, i, revin;
    
    if ((bfirdata_coeff_index(BFIR_FILTER_TYPE_REVERB, id, 0)) < 0) {
	return;
    }
    brutefir_settings.reverb_id = id;
    if (!isruntime()) {
        return;
    }
    cmd[0] = '\0';
    str = cmd;
    revin = brutefir_settings.channel_count[IN][VCU_CH_REVERB];
    for (n = 0; n < brutefir_settings.channel_count[OUT][VCU_CH_REVERB]; n++) {
        for (i = 0; i < revin; i++) {
            sprintf(str, "cfc %d %d;",
                    brutefir_settings.runtime_map.rev2filt[n][i],
                    bfirdata_coeff_index(BFIR_FILTER_TYPE_REVERB, id,
                                         revin * n + i));
            str += strlen(str);
        }
    }
    putcommand(cmd);    
}

void
bfir_set_equalisation(const uint32_t mask[],
		      int type,
		      int id,
                      int io)
{
    struct runtime_map *rmap;
    int n, coeff, filt;
    char cmd[20];

    if (type < 0 || type >= VCU_CH_WP_COUNT || (io != IN && io != OUT) ||
        mask == NULL)
    {
        elog(LOGDEBUG, "invalid param");
        return;
    }
    elog(LOGDEBUG, "mask[0] %08X, type %d, id %d, io %d",
         mask[0], type, id, io);
    
    if (id < 0) {
	coeff = -1;
	id = -1;
    } else {
        if ((coeff = bfirdata_coeff_index(BFIR_FILTER_TYPE_EQ, id, 0)) == -1) {
            elog(LOGDEBUG, "did not find filters for id %d", id);
	    return;
        }        
    }
    rmap = &brutefir_settings.runtime_map;
    n = -1;
    bfir_start_buffering_commands();
    while ((n = bit_find(mask, n + 1, BFIR_MAX_OUTPUT_CHANNELS - 1)) != -1) {
        if (!bit_isset(brutefir_settings.eq_active[io][type], n)) {
            continue;
        }
        brutefir_settings.eq_id[io][type][n] = id;
        if (!isruntime()) {
            continue;
        }
        if (type == VCU_CH_PHYSICAL) {
            filt = -1;
            while ((filt = bit_find(rmap->eq_phys2filt[io][n], filt + 1,
                                    rmap->filter_count - 1)) != -1)
            {
                sprintf(cmd, "cfc %d %d;", filt, coeff);
            }
        } else {
            sprintf(cmd, "cfc %d %d;", rmap->eq_logi2filt[io][type][n], coeff);
        }
        putcommand(cmd);
    }
    bfir_commit_buffered_commands();
}

void
bfir_set_xtalkcancel(const uint32_t mask[],
                     int id)
{
    char cmd[BFIR_MAX_OUTPUT_CHANNELS * 20], *str;
    int n, i;
    
    if (bfirdata_coeff_index(BFIR_FILTER_TYPE_XTC, id, 0) < 0) {
        elog(LOGDEBUG, "failed to set xtc to id %d", id);
	return;
    }
    cmd[0] = '\0';
    str = cmd;
    for (n = 0;
         n < brutefir_settings.channel_count[OUT][VCU_CH_AMBIOPOLE] / 2;
         n++)
    {
        if (bit_isset(mask, n) && brutefir_settings.xtalk_cancel[n]) {
            brutefir_settings.xtalkcancel_id[n] = id;
            if (isruntime()) {
                for (i = 0; i < 4; i++) {
                    sprintf(str, "cfc %d %d;",
                            brutefir_settings.runtime_map.xtc2filt[n][i],
                            bfirdata_coeff_index(BFIR_FILTER_TYPE_XTC, id, i));
                    str += strlen(str);
                }
            }
        }
    }
    if (isruntime()) {
        putcommand(cmd);
    }
    elog(LOGDEBUG, "set xtc to id %d", id);
}

void
bfir_set_dyneq(int id,
               const struct bfirdyneq *dyneq)
{
    char cmd[VCU_EQPRG_MAX_BANDS * 20 + 100], *str;
    struct bfirdyneq *cureq;
    int coeff, n, cureq_index;

    if (!isruntime() && id < 0 && dyneq == NULL) {
        elog(LOGDEBUG, "clear all dyneq");
        for (n = 0; n < brutefir_settings.dyneq_count; n++) {
            free(brutefir_settings.dyneq[n]);
            brutefir_settings.dyneq[n] = NULL;
        }
        brutefir_settings.dyneq_count = 0;
        return;
    }
    
    if ((coeff = bfirdata_coeff_index(BFIR_FILTER_TYPE_EQ, id, 0)) < 0) {
        elog(LOGDEBUG, "did not find filters for id %d", id);
        return;
    }
    cureq = NULL;
    cureq_index = -1;
    for (n = 0; n < brutefir_settings.dyneq_count; n++) {
        if (brutefir_settings.dyneq[n]->eq_id == id) {
            cureq = brutefir_settings.dyneq[n];
            cureq_index = n;
            break;
        }
    }
    if (isruntime()) {
        if (dyneq == NULL || cureq == NULL ||
            dyneq->n_bands != cureq->n_bands ||
            memcmp(dyneq->band, cureq->band,
                   cureq->n_bands * sizeof(double)) != 0)
        {
            elog(LOGDEBUG, "not valid runtime dyneq setting");
            return;
        }
        if (memcmp(dyneq->amp, cureq->amp,
                   cureq->n_bands * sizeof(double)) == 0)
        {
            elog(LOGDEBUG, "no change");
            return;
        }
        cmd[0] = '\0';
        str = cmd;
        sprintf(str, "lmc eq %d mag ", coeff);
        str += strlen(str);
        for (n = 0; n < cureq->n_bands; n++) {
            if (dyneq->amp[n] != cureq->amp[n]) {
                cureq->amp[n] = dyneq->amp[n];
                sprintf(str, "%.1f/%.2f,", dyneq->band[n], dyneq->amp[n]);
                str += strlen(str);
            }
        }
        str[-1] = ';';
        putcommand(cmd);
        return;
    }
    if (dyneq == NULL) {
        if (cureq == NULL) {
            elog(LOGDEBUG, "no change");
            return;
        }
        free(brutefir_settings.dyneq[cureq_index]);
        brutefir_settings.dyneq_count--;
        memmove(&brutefir_settings.dyneq[cureq_index],
                &brutefir_settings.dyneq[cureq_index + 1],
                (brutefir_settings.dyneq_count - cureq_index) *
                sizeof(struct bfirdyneq *));
        elog(LOGDEBUG, "removed dyneq");
        return;
    }
    if (cureq != NULL) {
        free(brutefir_settings.dyneq[cureq_index]);
    } else {
        cureq_index = brutefir_settings.dyneq_count;
        if (cureq_index == VCU_CH_COUNT * BFIR_MAX_OUTPUT_CHANNELS) {
            elog(LOGERROR, "dyneq table is full");
            return;
        }
    }
    if ((brutefir_settings.dyneq[cureq_index] =
         malloc(sizeof(struct bfirdyneq))) == NULL)
    {
        elog(LOGERROR, "memory allocation failed");
        return;
    }
    memcpy(brutefir_settings.dyneq[cureq_index], dyneq,
           sizeof(struct bfirdyneq));
    brutefir_settings.dyneq[cureq_index]->eq_id = id;
    brutefir_settings.dyneq_count++;
    elog(LOGDEBUG, "added/modified dyneq");
}

void
bfir_reset_peak_meter(void)
{
    if (isruntime()) {
        putcommand("rpk;");
    }
}

void
bfir_set_volume(int io,
                int type,
                const uint32_t mask[],
                int start,
                int end,
                double volume)
{
    uint32_t actumask[VCU_MAX_CHANNELS/32];
    int actu;

    if ((io != IN && io != OUT) ||
        type < 0 || type >= VCU_CH_WP_COUNT || mask == NULL)
    {
        elog(LOGERROR, "invalid parameter");
        return;
    }

    elog(LOGDEBUG, "io %d; type %d; mask[0] %08X; volume %f",
         io, type, mask[0], volume);
    
    memset(actumask, 0, sizeof(actumask));
    actu = start - 1;    
    while ((actu = bit_find(mask, actu + 1, end)) != -1) {
        if (volume == brutefir_settings.actu.ampl[io][type][actu]) {
            continue;
        }
        brutefir_settings.actu.ampl[io][type][actu] = volume;
        bit_set(actumask, actu);
    }
    apply_actu2virt_mdv(io, type, actumask, start, end);
}

void
bfir_set_delay(int io,
               int type,
               const uint32_t mask[],
               int start,
               int end,
               int delay)
{
    uint32_t actumask[VCU_MAX_CHANNELS/32];
    int actu;

    if ((io != IN && io != OUT) ||
        type < 0 || type >= VCU_CH_WP_COUNT || mask == NULL || delay < 0)
    {
        elog(LOGDEBUG, "io %d; type %d; mask %p; delay %d",
             io, type, mask, delay);
        elog(LOGERROR, "invalid parameter");
        return;
    }

    elog(LOGDEBUG, "io %d; type %d; mask[0] %08X; delay %d",
         io, type, mask[0], delay);
    
    memset(actumask, 0, sizeof(actumask));
    actu = start - 1;    
    while ((actu = bit_find(mask, actu + 1, end)) != -1) {
        if (delay == brutefir_settings.actu.delays[io][type][actu]) {
            continue;
        }
        brutefir_settings.actu.delays[io][type][actu] = delay;
        bit_set(actumask, actu);
    }
    apply_actu2virt_mdv(io, type, actumask, start, end);
}

void
bfir_set_active(int io,
                int type,
                const uint32_t mask[],
                int start,
                int end)
{
    uint32_t actumask[VCU_MAX_CHANNELS/32];
    int actu;

    if ((io != IN && io != OUT) ||
        type < 0 || type >= VCU_CH_WP_COUNT || mask == NULL)
    {
        elog(LOGERROR, "invalid parameter");
        return;
    }

    memset(actumask, 0, sizeof(actumask));
    actu = -1;
    for (actu = start; actu <= end; actu++) {
        if (!!bit_isset(mask, actu) !=
            !brutefir_settings.actu.muted[io][type][actu])
        {
            brutefir_settings.actu.muted[io][type][actu] =
                !bit_isset(mask, actu);
            bit_set(actumask, actu);            
        }
    }
    apply_actu2virt_mdv(io, type, actumask, start, end);
}

void
bfir_start_buffering_commands(void)
{
    if (buffer_commands == 0) {
        if (cmdbuffer == NULL) {
            cmdbuffer_cap = 100;
            cmdbuffer = malloc(100);
        }
        cmdbuffer[0] = '\0';
    }
    buffer_commands++;
}

void
bfir_commit_buffered_commands(void)
{
    buffer_commands--;
    if (buffer_commands == 0) {
        cmdbuffer_pos = 0;
        putcommand(cmdbuffer);
    }
}

void
bfir_stop(void)
{
    if (isruntime()) {
        putcommand("abort;");
    }
}

int
bfir_start(int sample_rate,
	   int maxdelay[2],
           bool_t powersave,
           bool_t watch_clock,
	   void (*peak_meter)(double peak),
	   void (*process_terminated)(int reason),
           int sample_format,
           int channel_count,
           bool_t loop,
           const char file_path[])
{
    int n, i, io, in_pipe[2], ri, status, fd, type, bsize, coeff, n_coeffs;
    int cli_in_pipe[2], cli_out_pipe[2];
    const struct brutefir_coeff *coeffs;
    char console_path[PATH_MAX + 100];
    struct brutefir_config config;
    char path[PATH_MAX + 100];
    FILE *stream;
    int count;

    MUST_NOT_BE_RUNNING(VCU_ERR_PREREQUISITES_NOT_FULFILLED);

    bfir_peak_meter = peak_meter;
    bfir_process_terminated = process_terminated;
    
    for (ri = 0; ri < hard_settings.n_rates; ri++) {
	if (hard_settings.rates[ri] == sample_rate) {
	    break;
	}
    }
    if (ri == hard_settings.n_rates) {
	elog(LOGDEBUG, "No matching sample rate found.");
	return VCU_ERR_UNSUPPORTED_SAMPLE_RATE;
    }

    memset(&config, 0, sizeof(config));
    
    config.sample_rate = hard_settings.rates[ri];
    config.block_length = bsize = hard_settings.block_sizes[ri];
    config.n_blocks = bfirdata_longest_filter_blocks(config.sample_rate);    
    config.n_processors = sysmon_n_cpus();
    config.show_progress = false;
    config.apply_dither = brutefir_settings.apply_dither;
    config.powersave = powersave;
    config.watch_clock = watch_clock;
    config.allow_poll_mode = brutefir_settings.allow_poll_mode;
    config.double_precision = !!(hard_settings.realsize > 4);
    strcpy(config.modules_path, brutefir_settings.modules_path);
    strcpy(config.convolver_config, brutefir_settings.convolver_config);
    memcpy(config.channel_count, brutefir_settings.channel_count,
           sizeof(config.channel_count));
    config.ambisonic_input_format = brutefir_settings.ambisonic_input_format;

    config.chmap = brutefir_settings.chmap;
    memcpy(config.ampl, brutefir_settings.virt.ampl, sizeof(config.ampl));
    memcpy(config.muted, brutefir_settings.virt.muted, sizeof(config.muted));
    memcpy(config.delays, brutefir_settings.virt.delays, sizeof(config.delays));
    memcpy(config.ext_channel_order, ext_channel_order,
           sizeof(ext_channel_order));
    FOR_IN_AND_OUT {
        config.device_count[io] = brutefir_settings.device_count[io];
        config.max_delay[io] = maxdelay[io];
        if (io == IN && file_path != NULL) {
            config.device_count[io] = 1;
            strcpy(config.device[io][0], "file");
            sprintf(config.device_params[io][0],
                    "{ path: \"%s\"; loop: %s; }", file_path,
                    loop ? "true" : "false");
            config.sample_format[io] = sample_format;
            config.open_channels[io][0] = channel_count;
            config.tot_open_channels[io] = channel_count;
        } else {
            config.tot_open_channels[io] = 0;
            config.sample_format[io] = brutefir_settings.sample_format[io];
            for (n = 0; n < config.device_count[io]; n++) {
                strcpy(config.device[io][n], brutefir_settings.device[io][n]);
                strcpy(config.device_params[io][n],
                       brutefir_settings.device_params[io][n]);
                config.open_channels[io][n] =
                    brutefir_settings.open_channels[io][n];
                config.tot_open_channels[io] +=
                    brutefir_settings.open_channels[io][n];
            }
        }
    }
    elog(LOGDEBUG, "%d/%d in/out devices",
         config.device_count[IN], config.device_count[OUT]);
    
    memcpy(config.xtalk_cancel, brutefir_settings.xtalk_cancel,
           sizeof(config.xtalk_cancel));

    if ((coeffs = bfirdata_get_coeffs(&n_coeffs, sample_rate)) == NULL) {
        elog(LOGERROR, "could not get brutefir coefficients");
        exit(EXIT_FAILURE);
    }
    
    /* equalisation */
    FOR_IN_AND_OUT {
        for (type = 0; type < VCU_CH_WP_COUNT; type++) {
            if (type != VCU_CH_PHYSICAL) {
                count = config.channel_count[io][type];
            } else {
                count = config.tot_open_channels[io];
            }
            for (n = 0; n < count; n++) {
                if (bit_isset(brutefir_settings.eq_active[io][type], n)) {
                    elog(LOGDEBUG, "eq active for %d %d %d", io, type, n);
                    bit_set(config.eqactive[io][type], n);
                    if (brutefir_settings.eq_id[io][type][n] < 0) {
                        config.eqcoeffs[io][type][n] = -1;
                    } else {
                        config.eqcoeffs[io][type][n] =
                            bfirdata_coeff_index(BFIR_FILTER_TYPE_EQ,
                                                 brutefir_settings.eq_id
                                                 [io][type][n], 0);
                        if (config.eqcoeffs[io][type][n] < 0) {
                            elog(LOGDEBUG, "eq %d not found",
                                 brutefir_settings.eq_id[io][type][n]);
                            return VCU_ERR_PREREQUISITES_NOT_FULFILLED;
                        }
                    }
                } else {
                    bit_clr(config.eqactive[io][type], n);
                    config.eqcoeffs[io][type][n]= -1;
                }
            }
        }
    }
    /* dynamic equalisation */
    elog(LOGDEBUG, "there are %d dynamic equalisers",
         brutefir_settings.dyneq_count);
    for (n = 0; n < brutefir_settings.dyneq_count; n++) {
        coeff = bfirdata_coeff_index(BFIR_FILTER_TYPE_EQ,
                                     brutefir_settings.dyneq[n]->eq_id, 0);
        elog(LOGDEBUG, "dyneq %d coeff is %d", n, coeff);
        if (coeff < 0) {
            elog(LOGDEBUG, "eq %d not found", coeff);
            return VCU_ERR_PREREQUISITES_NOT_FULFILLED;
        }
        FOR_IN_AND_OUT {
            for (type = 0; type < VCU_CH_WP_COUNT; type++) {
                if (type != VCU_CH_PHYSICAL) {
                    count = config.channel_count[io][type];
                } else {
                    count = config.tot_open_channels[io];
                }
                for (i = 0; i < count; i++) {
                    if (config.eqcoeffs[io][type][i] == coeff) {
                        goto post_dyneq_coeff_loop;
                    }
                }
            }
        }
    post_dyneq_coeff_loop:
        if (type == VCU_CH_WP_COUNT) {
            /* this equaliser is not used */
            elog(LOGDEBUG, "dyneq with eq id %d is not used",
                 brutefir_settings.dyneq[n]->eq_id);
            continue;
        }
        config.dyneq[config.dyneq_count] = alloca(sizeof(struct bfirdyneq));
        memcpy(config.dyneq[config.dyneq_count], brutefir_settings.dyneq[n],
               sizeof(struct bfirdyneq));
        for (i = 0; i < n_coeffs; i++) {
            if (coeffs[i].id == config.dyneq[config.dyneq_count]->eq_id) {
                break;
            }
        }
        if (i == n_coeffs) {
            elog(LOGERROR, "dyneq coeff not found, aborting");
            exit(EXIT_FAILURE);
        }
        config.dyneq[config.dyneq_count]->eq_id = coeff;
        config.dyneq[config.dyneq_count]->blocks = coeffs[i].n_blocks;
        config.dyneq_count++;
    }
    
    /* ambiopole stuff */
    for (n = 0; n < config.channel_count[OUT][VCU_CH_AMBIOPOLE] / 2; n++) {
        if (!brutefir_settings.xtalk_cancel[n]) {
            elog(LOGDEBUG, "no xtc for ambiopole %d", n);
            for (i = 0; i < 4; i++) {
                config.xtalkcancel_coeffs[4*n+i] = -1;
            }
            continue;
        }       
        for (i = 0; i < 4; i++) {
            config.xtalkcancel_coeffs[4*n+i] = 
                bfirdata_coeff_index(BFIR_FILTER_TYPE_XTC,
                                     brutefir_settings.xtalkcancel_id[n], i);
            if (config.xtalkcancel_coeffs[4*n+i] < 0) {
                elog(LOGDEBUG, "xtc filter %d not found",
                     brutefir_settings.xtalkcancel_id[n]);
                return VCU_ERR_PREREQUISITES_NOT_FULFILLED;
            }
        }
    }
    
    /* reverb stuff */
    if (config.channel_count[OUT][VCU_CH_REVERB] > 0) {
        for (n = 0;
             n < config.channel_count[IN][VCU_CH_REVERB] *
                 config.channel_count[OUT][VCU_CH_REVERB];
             n++)
        {
            config.reverb_coeffs[n] =
                bfirdata_coeff_index(BFIR_FILTER_TYPE_REVERB,
                                     brutefir_settings.reverb_id, n);
            if (config.reverb_coeffs[n] < 0) {
                elog(LOGDEBUG, "reverb filter %d not found",
                     brutefir_settings.reverb_id);
                return VCU_ERR_PREREQUISITES_NOT_FULFILLED;
            }
        }
    }

    /* ambisonic stuff */
    if (config.channel_count[OUT][VCU_CH_AMBISONIC] > 0) {
        memcpy(config.ambisonic_gains, brutefir_settings.ambisonic_gains,
               sizeof(config.ambisonic_gains));
    }

    if (pipe(in_pipe) == -1 ||
        pipe(cli_in_pipe) == -1 ||
        pipe(cli_out_pipe) == -1)
    {
        elog(LOGDEBUG, "Could not create pipe: %s.", strerror(errno));	
        return VCU_ERR_UNKNOWN_ERROR;
    }    
    /* the in_pipe[0] must not have consist of STDIN_FILENO, STDOUT_FILENO or
       STDERR_FILENO. This could happen if any of the FILENO has been closed */
    fd = fcntl(in_pipe[0], F_DUPFD, 3);
    close(in_pipe[0]);
    in_pipe[0] = fd;
    fd = fcntl(cli_in_pipe[0], F_DUPFD, 3);
    close(cli_in_pipe[0]);
    cli_in_pipe[0] = fd;
    fd = fcntl(cli_out_pipe[0], F_DUPFD, 3);
    close(cli_out_pipe[0]);
    cli_out_pipe[0] = fd;
    config.cli_pipe[0] = cli_in_pipe[0];
    config.cli_pipe[1] = cli_out_pipe[1];

    if (bfir_debug) {
        sprintf(console_path, "%s%s", log_path,
                BFIR_DEBUG_BRUTEFIR_CONSOLE_OUTPUT);
        if ((n = open(console_path, O_WRONLY | O_CREAT | O_TRUNC, 0666))
            == -1)
        {
            elog(LOGDEBUG, "Open \"%s\" failed: %s", console_path,
                 strerror(errno));
        }
    } else {
        if ((n = open(NULL_DEVICE, O_APPEND)) == -1) {
            elog(LOGDEBUG, "Open \"%s\" failed: %s", NULL_DEVICE,
                 strerror(errno));
        }
    }
    fd = fcntl(n, F_DUPFD, 3);
    close(n);
    
    elog(LOGDEBUG, "About to fork and exec \"%s\"", brutefir_settings.path);
    switch ((brutefir_pid = fork())) {
    case 0:
        close(cli_in_pipe[1]);
        close(cli_out_pipe[0]);
	close(in_pipe[1]);
	dup2(in_pipe[0], STDIN_FILENO);
	dup2(fd, STDOUT_FILENO);
	dup2(fd, STDERR_FILENO);
        bfirmem_finish();
        execl(brutefir_settings.path, brutefir_settings.path, "-nodefault",
              "stdin", NULL);
	elog(LOGERROR, "Exec failed: %s.", strerror(errno));
	exit(BF_EXIT_OTHER);
	break;
    case -1:
	elog(LOGERROR, "Fork failed: %s.", strerror(errno));
	return VCU_ERR_UNKNOWN_ERROR;
    default:
	break;
    }
    close(cli_in_pipe[0]);
    close(cli_out_pipe[1]);
    close(in_pipe[0]);
    close(fd);

    stream = fdopen(in_pipe[1], "w");
    if (!bfirconf_config_write(stream, &config, coeffs, n_coeffs,
                               &brutefir_settings.runtime_map))
    {
	fclose(stream);
	kill(brutefir_pid, SIGTERM);
	wait(&status);
	elog(LOGDEBUG, "Could not write config.");
        /* assume too many filters */
	return VCU_ERR_CONVOLVER_FAILED_TOO_COMPLEX;
    }	
    fclose(stream);
    if (bfir_debug) {
        sprintf(path, "%s%s", log_path, BFIR_DEBUG_LAST_BRUTEFIR_CONFIG);
	if ((stream = fopen(path, "wt")) != NULL) {
	    bfirconf_config_write(stream, &config, coeffs, n_coeffs,
                                  &brutefir_settings.runtime_map);
	    fclose(stream);
            elog_concatenate_file(path);
	}
    }

    brutefir_instream = fdopen(cli_in_pipe[1], "w");
    brutefir_outstream = fdopen(cli_out_pipe[0], "r");
    setvbuf(brutefir_instream, NULL, _IOLBF, 0);
    setvbuf(brutefir_outstream, NULL, _IOLBF, 0);

    putcommand("tp;upk;ppk;");

    return cli_out_pipe[0];
}

void
bfir_process(void)
{
    static int debug_lines = 0;

    char str[1000], *p, *p1, path[PATH_MAX + 100];
    int status, ch;
    float peak, f;
    FILE *stream;

    MUST_BE_RUNNING( );

    if ((fgets(str, 999, brutefir_outstream)) == NULL) {
	/* assume BruteFIR has died */
        elog(LOGDEBUG, "assuming BruteFIR has died");
        fclose(brutefir_instream);
	fclose(brutefir_outstream);
	brutefir_instream = NULL;
	brutefir_outstream = NULL;
	kill(brutefir_pid, SIGTERM); /* just in case */
	waitpid(brutefir_pid, &status, 0);
        if (bfir_debug) {
            sprintf(path, "%s%s", log_path, BFIR_DEBUG_BRUTEFIR_CONSOLE_OUTPUT);
            elog_concatenate_file(path);
        }
	bfir_process_terminated(translate_exitcode(WEXITSTATUS(status)));
	return;
    }
    str[999] = '\0';
    if (bfir_debug) {
        sprintf(path, "%s%s", log_path, BFIR_DEBUG_BRUTEFIR_CLI_OUTPUT);
        if ((stream = fopen(path, "at")) != NULL) {
	    fprintf(stream, "%d: %s", ++debug_lines, str);
	    fclose(stream);
	}
    }

    /* Note: ok reply is no longer used in BruteFIR... */
    if (str[0] == 'o' && str[1] == 'k') {
	return;
    }
    
    if ((p1 = strstr(str, "peak: ")) != NULL) {
	peak = VCU_PEAK_NEG_INFINITE;
	ch = 0;
	while ((p1 = strchr(p1, '.')) != NULL) {
	    p = p1++;
	    while (*p != '/' && p != str) p--;
            p++;
            if (p[0] == '-' && (p[1] == 'I' || p[1] == 'i')) {
                continue;
            }
	    f = atof(p);
	    if (f > peak) {
		peak = f;
	    }
	    ch++;
	}
	bfir_peak_meter(peak);
    }    
}

int
bfir_benchmark(int sample_rate,
               bool_t test_low_latency,
               bool_t apply_dither,
               int ambisonic_input_format,
               const struct bfir_chmap *maps[2],
	       double max_load,
	       double max_seconds,
	       int *n_alternatives,
	       int block_sizes[],
	       int n_blocks[],
               void (*filter_lengths)(int block_size,
                                      int sample_rate,
                                      int xtc_length[],
                                      int *eq_length[2][VCU_CH_WP_COUNT]),
	       void (*progress_callback)(int n_blocks,
					 int block_size,
					 double load))
{
    int n, i, io, reverb_length, max_block_size, start_block_size, type;
    int add_blocks, max_blocks, start_reverb_blocks;
    int block_size, n_reverb_blocks;
    double max_rti, rti, rti_low_bound, rti_high_bound, seconds;
    bool_t has_overshot, has_undershot, memory_limited, last_limited_by_memory;
    int n_xtalk_blocks[BFIR_MAX_OUTPUT_CHANNELS];
    int _n_eq_blocks[2][VCU_CH_WP_COUNT][BFIR_MAX_OUTPUT_CHANNELS];
    int _eq_length[2][VCU_CH_WP_COUNT][BFIR_MAX_OUTPUT_CHANNELS];
    int *eq_length[2][VCU_CH_WP_COUNT];
    int xtc_length[BFIR_MAX_OUTPUT_CHANNELS];
    const int *n_eq_blocks[2][VCU_CH_WP_COUNT];
    int max_reverb_lengths[BFIR_BENCHMARK_N_BLOCK_SIZES];
    int channel_count[2][VCU_CH_COUNT], count;
    int bf_exit_status, n_failed_runs;
    struct chmap chmap;
    void *dummy_ram;
    
    MUST_NOT_BE_RUNNING(BFIR_BENCHMARK_STATUS_INVALID_VALUE);    
    
    max_rti = max_load;
    rti_low_bound = max_rti * 0.95;
    rti_high_bound = max_rti;
    *n_alternatives = 0;
    add_blocks = 0;
    FOR_IN_AND_OUT {
        for (type = 0; type < VCU_CH_WP_COUNT; type++) {
            n_eq_blocks[io][type] = _n_eq_blocks[io][type];
            eq_length[io][type] = _eq_length[io][type];
        }
    }

    if (build_maps(maps, ambisonic_input_format, &chmap, channel_count) != 0) {
        elog(LOGDEBUG, "failed to build maps");
        return BFIR_BENCHMARK_STATUS_INVALID_VALUE;
    }
    
    for (type = 0; type < VCU_CH_COUNT && channel_count[IN][type] == 0; type++);
    if (type == VCU_CH_COUNT) {
        elog(LOGDEBUG, "there are no channels");
        /* there are no channels */
	return BFIR_BENCHMARK_STATUS_INVALID_VALUE;
    }
    
    /* we allocate 20 megabytes of RAM here, so we get BF_EXIT_NO_MEMORY a bit
       earlier so we have a marigin later on */
    dummy_ram = malloc(20 * 1024 * 1024);
    memset(dummy_ram, 0, 20 * 1024 * 1024);

    /* free all filter memory by calling bfir_hard_settings again */
    bfir_hard_settings(hard_settings.n_rates, hard_settings.rates,
                       hard_settings.block_sizes,
                       brutefir_settings.channel_count[IN][VCU_CH_REVERB],
                       brutefir_settings.channel_count[OUT][VCU_CH_REVERB],
                       brutefir_settings.ambisonic_input_format,
                       hard_settings.realsize);
    
    memset(max_reverb_lengths, 0, sizeof(max_reverb_lengths));
    n = sample_rate / 48001 + 1;
    max_block_size = n * BFIR_MAX_BLOCKSIZE_48000HZ;    
    if (test_low_latency) {
        block_size = start_block_size = BFIR_MIN_BLOCKSIZE_LOWLAT;
    } else {
        block_size = start_block_size = n * BFIR_MIN_BLOCKSIZE_48000HZ;
    }
    if (block_size < hardware_min_blocksize) {
        block_size = start_block_size = hardware_min_blocksize;
    }
    reverb_length = 0;
    
    last_limited_by_memory = false;
    n_failed_runs = 0;
    for (n = 0; block_size <= max_block_size; block_size *= 2, n++) {

        elog(LOGDEBUG, "block size: %d", block_size);
        
        filter_lengths(block_size,
                       sample_rate,
                       xtc_length,
                       eq_length);
        
	start_reverb_blocks = n_reverb_blocks = reverb_length / block_size + 1;

	/* calculate the number of blocks per filter with the current block
	   size */
        max_blocks = 0;
        FOR_IN_AND_OUT {
            for (type = 0; type < VCU_CH_WP_COUNT; type++) {
                if (type != VCU_CH_PHYSICAL) {
                    count = channel_count[io][type];
                } else {
                    count = brutefir_settings.tot_open_channels[io];
                }
                for (i = 0; i < count; i++) {
                    _n_eq_blocks[io][type][i] =
                        (eq_length[io][type][i] + block_size - 1) / block_size;
                    if (_n_eq_blocks[io][type][i] > max_blocks) {
                        max_blocks = n_eq_blocks[io][type][i];
                    }
                }
            }
        }
        elog(LOGDEBUG, "max blocks after eq: %d", max_blocks);
        for (i = 0; i < channel_count[OUT][VCU_CH_AMBIOPOLE] / 2; i++) {
            n_xtalk_blocks[i] = (xtc_length[i] + block_size - 1) / block_size;
            if (n_xtalk_blocks[i] > max_blocks) {
                max_blocks = n_xtalk_blocks[i];
            }
        }
        elog(LOGDEBUG, "max blocks after xtc: %d", max_blocks);
	if (max_blocks > BFIR_MAX_BLOCKS) {
	    /* too many blocks, try again (with increased block size) */
            elog(LOGDEBUG, "too many blocks");
	    continue;
	}

	if (channel_count[OUT][VCU_CH_REVERB] == 0) {
	    /* we want as short block size as possible, that is, we break as
	       soon as we can */
            elog(LOGDEBUG, "calling benchrun with block size %d", block_size);
	    if (!benchrun(sample_rate,
                          apply_dither,
			  block_size,
			  n_reverb_blocks,
			  n_xtalk_blocks,
			  n_eq_blocks,
                          ambisonic_input_format,
			  maps,
			  &bf_exit_status,
			  &rti))
	    {
		free(dummy_ram);
                elog(LOGDEBUG, "benchrun failed");
                return BFIR_BENCHMARK_STATUS_RUN_FAILED;
	    }

	    if (rti == 0.0 || rti > max_rti) {
		n_failed_runs++;
	    } else {
		n_failed_runs = 0;
	    }
	    if (n_failed_runs >= 5 && *n_alternatives == 0) {
		free(dummy_ram);
                elog(LOGDEBUG, "too demanding configuration");
		return BFIR_BENCHMARK_STATUS_OUT_OF_RANGE;
	    }
	    switch (bf_exit_status) {
	    case BF_EXIT_OK:
		break;
	    case BF_EXIT_NO_MEMORY:
                elog(LOGDEBUG, "no memory");
		free(dummy_ram);
                return BFIR_BENCHMARK_STATUS_OUT_OF_RANGE;
	    default:
                elog(LOGDEBUG, "run failed (exit status %d)", bf_exit_status);
		free(dummy_ram);
		return BFIR_BENCHMARK_STATUS_RUN_FAILED;
	    }
	    if (progress_callback != NULL) {
		elog(LOGDEBUG, "progress_callback: X %d %f", block_size, rti);
		progress_callback(0, block_size, rti);
	    }
	    if (rti <= rti_high_bound) {
		block_sizes[0] = block_size;
		n_blocks[0] = max_blocks;
		*n_alternatives = 1;
		free(dummy_ram);
                if (n == 0) {
                    elog(LOGDEBUG, "hard limited");
                    return BFIR_BENCHMARK_STATUS_HARD_LIMITED;
                }
                elog(LOGDEBUG, "processor limited");
                return BFIR_BENCHMARK_STATUS_PROCESSOR_LIMITED;
	    } else {
		/* try again */
                elog(LOGDEBUG, "try again");
		continue;
	    }
	}

	has_undershot = has_overshot = false;
	while (n_reverb_blocks > 0) {
	    memory_limited = false;
	    while ((seconds = (double)(n_reverb_blocks * block_size) /
		    (double)sample_rate) > max_seconds)
	    {
		memory_limited = true;
		if (n_reverb_blocks == 1) {
		    break;
		}
		n_reverb_blocks--;
	    }
	    if (seconds > max_seconds ||
		n_reverb_blocks < start_reverb_blocks)
	    {
		free(dummy_ram);
		if (*n_alternatives == 0) {
                    elog(LOGDEBUG, "out of range");
		    return BFIR_BENCHMARK_STATUS_OUT_OF_RANGE;
		}
                if (seconds > max_seconds || last_limited_by_memory) {
                    elog(LOGDEBUG, "memory limited");
                    return BFIR_BENCHMARK_STATUS_MEMORY_LIMITED;
                }
                elog(LOGDEBUG, "processor limited");
                return BFIR_BENCHMARK_STATUS_PROCESSOR_LIMITED;
	    }
	
            elog(LOGDEBUG, "calling benchrun with block size %d", block_size);
	    if (!benchrun(sample_rate,
                          apply_dither,
			  block_size,
			  n_reverb_blocks,
			  n_xtalk_blocks,
			  n_eq_blocks,
                          ambisonic_input_format,
			  maps,
			  &bf_exit_status,
			  &rti))
	    {
                elog(LOGDEBUG, "benchrun failed");
		free(dummy_ram);
		return BFIR_BENCHMARK_STATUS_RUN_FAILED;
	    }
            elog(LOGDEBUG, "benchmark run: rti: %f, status %d",
                 rti, bf_exit_status);
	    if (rti == 0.0 || rti > max_rti) {
		n_failed_runs++;
	    } else {
		n_failed_runs = 0;
	    }
	    if (n_failed_runs >= 5 && *n_alternatives == 0) {
                elog(LOGDEBUG, "too demanding configuration");
		/* the configuration is too demanding for us */
		free(dummy_ram);
		return BFIR_BENCHMARK_STATUS_OUT_OF_RANGE;
	    }
	    switch (bf_exit_status) {
	    case BF_EXIT_OK:
		break;
	    case BF_EXIT_NO_MEMORY:
		/* here rti is 0.0 */
		break;
	    default:
                elog(LOGDEBUG, "run failed (exit status %d)", bf_exit_status);
		free(dummy_ram);
		return BFIR_BENCHMARK_STATUS_RUN_FAILED;
	    }
	    if (progress_callback != NULL) {
		elog(LOGDEBUG, "progress_callback: %d %d %f",
		     n_reverb_blocks, block_size, rti);
		progress_callback(n_reverb_blocks, block_size, rti);
	    }
	    if (rti == 0.0) {
		rti = max_rti + 1; /* just some number larger than max_rti */
	    }            
	    if ((rti >= rti_low_bound && rti <= rti_high_bound) ||
		(rti <= rti_high_bound && memory_limited) ||
		(rti <= rti_high_bound && add_blocks == 1 && has_overshot))
	    {
		elog(LOGDEBUG, "found alternative %f %f %f %d %d %d",
		     rti, rti_low_bound, rti_high_bound, add_blocks,
		     has_overshot, memory_limited);
		break;
	    }
	    if (rti < max_rti) {
		has_undershot = true;
		if (!has_overshot) {
		    add_blocks = n_reverb_blocks;
		} else {
		    add_blocks /= 2;
		}
		if (add_blocks == 0) {
		    add_blocks = 1;
		}
		n_reverb_blocks += add_blocks;
	    } else {
		has_overshot = true;
		if (!has_undershot) {
		    add_blocks = n_reverb_blocks / 2;
		} else {
		    add_blocks /= 2;
		}
		if (add_blocks == 0) {
		    add_blocks = 1;
		}
		n_reverb_blocks -= add_blocks;
	    }
	}

	if (n_reverb_blocks > 0) {
	    reverb_length = n_reverb_blocks * block_size;
	    if (*n_alternatives == 0 ||
		reverb_length > max_reverb_lengths[*n_alternatives - 1])
	    {
                last_limited_by_memory = rti < rti_low_bound;
		block_sizes[*n_alternatives] = block_size;
		n_blocks[*n_alternatives] = n_reverb_blocks;
		max_reverb_lengths[*n_alternatives] = reverb_length;
		(*n_alternatives)++;
	    }
	}
    }
    
    free(dummy_ram);
    if (*n_alternatives == 0) {
        elog(LOGDEBUG, "out of range");
        return BFIR_BENCHMARK_STATUS_OUT_OF_RANGE;
    }
    if (last_limited_by_memory) {
        elog(LOGDEBUG, "memory limited");
        return BFIR_BENCHMARK_STATUS_MEMORY_LIMITED;
    }
    elog(LOGDEBUG, "processor limited");
    return BFIR_BENCHMARK_STATUS_PROCESSOR_LIMITED;
}
