/*
 * (c) Copyright 2002 - 2003, 2006 -- Anders Torger
 *
 * This program is open source. For license terms, see the LICENSE file.
 *
 */
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <sys/ipc.h>
#include <sys/shm.h>

#include "defs.h"
#include "elog.h"
#include "bfirmem.h"

/* These limit is typically configurable, we use some standard fairly
   small value here, but not too small in order to avoid too many
   segments */
#ifndef SHMMAX
#define SHMMAX 0x2000000
#endif

struct allocmem {
    int offset;
    int size;
};

struct shmseg {
    struct allocmem *allocated;
    int n_allocated;
    int size;
    int id;
    uint8_t *p;
};

static struct shmseg *shmsegs;
static int n_shmsegs;
static int bytes_left;
static int bytes_total;

static void
free_block(struct shmseg *shmseg,
           const struct allocmem *allocmem)
{
    int n;
    
    for (n = 0; n < shmseg->n_allocated; n++) {
        if (memcmp(&shmseg->allocated[n], allocmem,
                   sizeof(struct allocmem)) == 0)
        {
            break;
        }
    }
    if (n == shmseg->n_allocated) {
        elog(LOGERROR, "block mismatch");
        exit(EXIT_FAILURE);
    }
    shmseg->n_allocated--;
    if (n != shmseg->n_allocated) {
        memmove(&shmseg->allocated[n], &shmseg->allocated[n+1],
                (shmseg->n_allocated - n) * sizeof(struct allocmem));
    }
}

static void
allocate_block(struct shmseg *shmseg,
               const struct allocmem *allocmem)
{
    struct allocmem *allocated;
    int n;
    
    allocated = realloc(shmseg->allocated, (shmseg->n_allocated + 1) *
                        sizeof(struct allocmem));
    if (allocated == NULL) {
        elog(LOGERROR, "out of memory, realloc failed");
        exit(EXIT_FAILURE);
    }
    shmseg->allocated = allocated;
    for (n = 0; n < shmseg->n_allocated; n++) {
        if (shmseg->allocated[n].offset > allocmem->offset) {
            break;
        }
    }
    if (n != shmseg->n_allocated) {
        memmove(&shmseg->allocated[n+1], &shmseg->allocated[n],
                (shmseg->n_allocated - n) * sizeof(struct allocmem));
    }
    shmseg->allocated[n] = *allocmem;
    shmseg->n_allocated++;
}

static void
get_largest_free_block(struct shmseg *shmseg,
                       struct allocmem *allocmem)
{
    int n, size, offset;
    
    if (shmseg->n_allocated == 0) {
        allocmem->offset = 0;
        allocmem->size = shmseg->size;
        return;
    }
    allocmem->offset = 0;
    allocmem->size = 0;
    offset = 0;
    size = 0;
    for (n = 0; n < shmseg->n_allocated; n++) {
        size = shmseg->allocated[n].offset - offset;
        if (size > allocmem->size) {
            allocmem->size = size;
            allocmem->offset = offset;
        }
        offset = shmseg->allocated[n].offset + shmseg->allocated[n].size;
    }
    size = shmseg->size - offset;
    if (size > allocmem->size) {
        allocmem->size = size;
        allocmem->offset = offset;
    }
}

static int
alloc_shared_memory(void **p,
		    size_t size)
{
    static key_t key = 1;
    struct shmid_ds shmid_ds;
    int n;

    while ((n = shmget(key, size,
                       IPC_CREAT | IPC_EXCL | 0666)) == -1 &&
           errno == EEXIST)
    {
        key++;
    }
    if (n == -1) {
	elog(LOGERROR, "Failed to allocate shared memory: %s.",
             strerror(errno));
	return -1;
    }
    
    if  ((*p = shmat(n, 0, 0)) == (char *)-1) {
	elog(LOGERROR, "Failed to attach to shared memory (shmid %d): %s.",
	     n, strerror(errno));
	return -1;	
    }
    memset(&shmid_ds, 0, sizeof(shmid_ds));
    if (shmctl(n, IPC_RMID, &shmid_ds) == -1) {
	elog(LOGERROR, "Failed to set IPC_RMID (shmid %d): %s.", n,
             strerror(errno));
	return -1;		
    }
    
    key++;

    memset(*p, 0, size);
    
    return n;
}

bool_t
bfirmem_init(int alloc_bytes)
{
    void *p;
    int n;

    n_shmsegs = (alloc_bytes + SHMMAX - 1) / SHMMAX;

    bytes_left = alloc_bytes;
    bytes_total = alloc_bytes;
    shmsegs = malloc(n_shmsegs * sizeof(struct shmseg));
    memset(shmsegs, 0, n_shmsegs * sizeof(struct shmseg));
    for (n = 0; alloc_bytes > 0; n++) {	
	if (alloc_bytes > SHMMAX) {
	    shmsegs[n].size = SHMMAX;
	} else {
	    shmsegs[n].size = alloc_bytes;
	}
        p = NULL;
	shmsegs[n].id = alloc_shared_memory(&p, shmsegs[n].size);
	if (shmsegs[n].id == -1) {
	    return false;
	}
        shmsegs[n].p = p;
	alloc_bytes -= shmsegs[n].size;
    }
    return true;    
}

void
bfirmem_finish(void)
{
    int n;

    for (n = 0; n < n_shmsegs; n++) {
        shmdt(shmsegs[n].p);
    }
    free(shmsegs);
    n_shmsegs = 0;
}

bool_t
bfirmem_alloc(struct memseg *memseg,
	      int alloc_bytes,
	      int block_size)
{
    struct allocmem allocmem;
    int n, i, bytes;

    if (alloc_bytes % block_size != 0) {
	elog(LOGDEBUG, "alloc_bytes not divisable with block_size");
	return false;
    }
    bytes = alloc_bytes;
    
    memset(memseg, 0, sizeof(struct memseg));
    for (n = i = 0; n < n_shmsegs && alloc_bytes != 0; n++) {
        get_largest_free_block(&shmsegs[n], &allocmem);        
        while ((allocmem.size = allocmem.size / block_size * block_size) > 0 &&
               alloc_bytes > 0 &&
               memseg->n_segments < BFIR_MAX_SEGMENTS)
        {
            memseg->segments[i] = shmsegs[n].p + allocmem.offset;
            memseg->segids[i] = shmsegs[n].id;
            memseg->segoffsets[i] = allocmem.offset;
            if (allocmem.size > alloc_bytes) {
                allocmem.size = alloc_bytes;
            }
            allocate_block(&shmsegs[n], &allocmem);
            memseg->seglengths[i] = allocmem.size;
            alloc_bytes -= allocmem.size;
            i++;
            memseg->n_segments = i;
            get_largest_free_block(&shmsegs[n], &allocmem);        
        }
    }
    if (alloc_bytes != 0) {
	elog(LOGDEBUG, "out of shared memory");
        bytes_left -= bytes - alloc_bytes;
        bfirmem_free(memseg);
	return false;
    }
    bytes_left -= bytes;
    elog(LOGDEBUG, "bytes_left: %d, alloc_bytes: %d, (segid %d)",
         bytes_left, bytes, memseg->segids[0]);
    return true;
}

void
bfirmem_free(struct memseg *memseg)
{
    struct allocmem allocmem;
    int n, i, bytes;

    i = n = bytes = 0;
    while (i < memseg->n_segments) {
        if (memseg->segments[i] >= shmsegs[n].p &&
            memseg->segments[i] < shmsegs[n].p + shmsegs[n].size)
        {
            allocmem.offset = memseg->segoffsets[i];
            allocmem.size = memseg->seglengths[i];
            free_block(&shmsegs[n], &allocmem);
            bytes += memseg->seglengths[i];
            i++;
        } else {
            n++;
            if (n == n_shmsegs) {
                elog(LOGERROR, "invalid free");
                return;
            }
        }
    }
    bytes_left += bytes;
    elog(LOGDEBUG, "bytes_left: %d, free_bytes: %d (segid %d)",
         bytes_left, bytes, memseg->segids[0]);
}

void
bfirmem_freeall(void)
{
    int n;
    
    for (n = 0; n < n_shmsegs; n++) {
	shmsegs[n].n_allocated = 0;
        free(shmsegs[n].allocated);
        shmsegs[n].allocated = NULL;
    }
    bytes_left = bytes_total;
}

int
bfirmem_bytes_left(void)
{
    return bytes_left;
}
