/*
 * (c) Copyright 2002 - 2004, 2006 -- Anders Torger
 *
 * This program is open source. For license terms, see the LICENSE file.
 *
 */
#include <string.h>
#include <math.h>

#include "alignalloc.h"
#include "bfirfilter.h"
#include "equaliser.h"
#include "almusvcu.h"
#include "convolve.h"
#include "reverb.h"
#include "decorr.h"
#include "elog.h"

struct stage {
    double azimuth;
    double elevation;
    double width;
    double height;
};

struct {
    double bformat_directivity;
    double early_length;
    double direct_length;
    double dry_level;
    bool_t force_delay_diff;
    bool_t force_decay;
    int response_mode;
    int decorrelate_mode;
    int n_sources;
    int stage_count;
    struct stage *stage;
} settings;

struct direction {
    int index;
    double az;
    double el;
};

static bool_t
within_stage(double azimuth,
             double elevation)
{
    int n;

    for (n = 0; n < settings.stage_count; n++) {
        if (azimuth >= settings.stage[n].azimuth -
            settings.stage[n].width / 2 &&
            azimuth <= settings.stage[n].azimuth +
            settings.stage[n].width / 2 &&
            elevation >= settings.stage[n].elevation -
            settings.stage[n].height / 2 &&
            elevation <= settings.stage[n].elevation +
            settings.stage[n].height / 2)
        {
            return true;
        }
    }
    
    return false;
}

static double
anglediff(double az1,
          double el1,
          double az2,
          double el2)
{
    double ux, uy, uz, vx, vy, vz, arc;

    ux = cos(el1 / 180.0 * M_PI) * cos(az1 / 180.0 * M_PI);
    uy = cos(el1 / 180.0 * M_PI) * sin(az1 / 180.0 * M_PI);
    uz = sin(el1 / 180.0 * M_PI);

    vx = cos(el2 / 180.0 * M_PI) * cos(az2 / 180.0 * M_PI);
    vy = cos(el2 / 180.0 * M_PI) * sin(az2 / 180.0 * M_PI);
    vz = sin(el2 / 180.0 * M_PI);

    arc = (acos(ux * vx + uy * vy + uz * vz) * 180.0 / M_PI);
    return arc;
}

static int
anglecmp(const void *a,
         const void *b)
{
    double diff1, diff2;
    
    diff1 = anglediff(0, 0, ((const struct direction *)a)->az,
                      ((const struct direction *)a)->el);
    diff2 = anglediff(0, 0, ((const struct direction *)b)->az,
                      ((const struct direction *)b)->el);

    if (diff1 < diff2) {
        return -1;
    } else if (diff1 > diff2) {
        return 1;
    }
    return 0;
}

static bool_t
isback(double azimuth,
       double range)
{
    return ((azimuth <= -90.0 + (range / 2.0 - 90.0)) ||
            (azimuth >= 90.0 - (range / 2.0 - 90.0)));
}

static int
bestreceiver(double azimuth,
             double elevation,
             const double receiver_azimuth[],
             const double receiver_elevation[],
             const bool_t used[],
             int n_receivers)
{
    double mindiff, recv_azimuth, diff;
    bool_t one_sided;
    int n, minpos;
    
    /* do we have one-sided receivers (that is on the left side only)? */
    one_sided = true;
    for (n = 0; n < n_receivers; n++) {
        if (receiver_azimuth[n] < 0) {
            one_sided = false;
            break;
        }
    }
    
    minpos = 0;
    mindiff = 10000.0;
    for (n = 0; n < n_receivers; n++) {
        if (used != NULL && used[n]) {
            continue;
        }
        if (one_sided && azimuth < 0) {
            recv_azimuth = -receiver_azimuth[n];
        } else {
            recv_azimuth = receiver_azimuth[n];
        }
        diff = anglediff(recv_azimuth,
                         receiver_elevation[n],
                         azimuth,
                         elevation);
        
        if (isback(azimuth, 200)) {
            /* rather assign back responses to the sides than front
               ones. If there are only front responses, we don't care */
            if (isback(recv_azimuth, 180)) {
                if (!isback(receiver_azimuth[minpos], 180)) {
                    mindiff = diff;
                    minpos = n;
                } else {
                    goto standard_select;
                }
            } else if (!isback(receiver_azimuth[minpos], 180)) {
                goto standard_select;
            }
        } else {
        standard_select:
            if (diff == mindiff) {
                /* if same difference, pick response farther to the back */
                if (fabs(recv_azimuth) > fabs(receiver_azimuth[minpos])) {
                    minpos = n;
                }
            } else if (diff < mindiff) {
                mindiff = diff;
                minpos = n;
            }
        }
    }
    if (mindiff == 10000.0) {
        return -1;
    }
    return minpos;
}

static double
get_max_db(void *imp,
           int length,
           int start,
           int window_size,
           int realsize)
{
    double max_db, db;
    int n;

    if (start < 0) {
        window_size += start;
        start = 0;
    }
    if (start > length) {
        start = length;
    }
    if (start + window_size > length) {
        window_size = length - start;
    }
    if (window_size < 0) {
        window_size = 0;
    }
    
    max_db = -150.0;
    if (realsize == 4) {
        for (n = start; n < start + window_size; n++) {
            if (((float *)imp)[n] != 0.0 &&
                (db = 20 * log10(fabsf(((float *)imp)[n]))) > max_db)
            {
                max_db = db;
            }
        }
    } else {
        for (n = start; n < start + window_size; n++) {
            if (((double *)imp)[n] != 0.0 &&
                (db = 20 * log10(fabs(((double *)imp)[n]))) > max_db)
            {
                max_db = db;
            }
        }
    }
    return max_db;
}

static int
calc_reverb_time(void *imp,
                 int length,
                 int predelay,
                 int early_length,
                 double db_decay,
                 double init_db,
                 int realsize)
{
    double db;
    int start;

    if (length <= early_length) {
        return length;
    }
    start = 0;
    db = get_max_db(imp, length, start, early_length, realsize);
    if (predelay > 0) {
        if (init_db <= -200.0) {
            init_db = 0.0;
        }
        start = -early_length;
    } else {
        init_db = db;
    }
    for (start += early_length;
         start + early_length <= length;
         start += early_length)
    {
        db = get_max_db(imp, length, start, early_length, realsize);
        if (init_db - db >= db_decay) {
            break;
        }
    }
    if (init_db - db < db_decay) {
        elog(LOGDEBUG, "decay too small: %fdB", init_db - db);
        return 0;
    }
    if (start + early_length > length) {
        return length + predelay;
    }
    return start + early_length + predelay;
}

static double
calc_tail_volume_prepf(const struct prepfilter *prepf)
{
    int n, ri, early_length;
    double tail_vol;

    tail_vol = 0.0;
    for (n = 0; n < prepf->n_channels; n++) {
        for (ri = 0; ri < prepf->n_rates; ri++) {
            early_length = (int)(settings.early_length *
                                 (double)prepf->rates[ri]);
            tail_vol += get_max_db(prepf->filters[n][ri], prepf->lengths[ri],
                                   prepf->lengths[ri] - early_length,
                                   early_length, prepf->realsize);
        }
    }
    tail_vol /= (double)(prepf->n_channels * prepf->n_rates);
    return tail_vol;
}                       

static double
calc_reverb_time_prepf(const struct prepfilter *prepf,
                       double predelay,
                       double db_decay)
{
    int n, int_rt60, ri;
    double rt60;

    rt60 = 0.0;
    for (n = 0; n < prepf->n_channels; n++) {
        for (ri = 0; ri < prepf->n_rates; ri++) {
            int_rt60 =
                calc_reverb_time(prepf->filters[n][ri],
                                 prepf->lengths[ri],
                                 (int)(predelay * (double)prepf->rates[ri]),
                                 (int)(settings.early_length *
                                       (double)prepf->rates[ri]),
                                 db_decay,
                                 -1000.0,
                                 prepf->realsize);
            rt60 += (double)int_rt60 / (double)prepf->rates[ri];
        }
    }
    rt60 /= (double)(prepf->n_channels * prepf->n_rates);
    return rt60;
}

/*
static void
flatten_reverb_tail(real_t imp[],
                    int length,
                    int early_length)
{
    real_t startdb, enddb, scale, r;
    int n;

    if (length < 4 * early_length) {
        elog(LOGDEBUG, "too short to flatten");
        return;
    }
    startdb = get_max_db(imp, early_length, early_length / 4);
    enddb = get_max_db(imp, 3 * early_length, early_length / 4);
    if (startdb <= enddb) {
        elog(LOGDEBUG, "startdb (%f) <= enddb (%f)", startdb, enddb);
        return;
    }
    for (n = early_length; n < 3 * early_length; n++) {
        r = (real_t)(n - early_length) / (real_t)(2 * early_length);
        scale = pow(10, (1.0 - r) * (enddb - startdb) / 20.0);
        imp[n] *= scale;
    }
}
*/

static void
shorten_reverb(void *imp,
               int length,
               int early_length,
               int target_length,
               int realsize)
{
    double db, tail_db, target_db;
    int n;
    
    if (target_length <= early_length || length <= target_length) {
        return;
    }

    /* calculate end dB (max of 'early_length' samples) */
    target_db = get_max_db(imp, length, length - early_length, early_length,
                           realsize);
    elog(LOGDEBUG, "target volume: %f dB", target_db);

    /* calculate current end dB (max of 'early_length' samples) */
    tail_db = get_max_db(imp, length, target_length - early_length,
                         early_length, realsize);
    elog(LOGDEBUG, "current tail volume: %f dB", tail_db);

    /* linearly (with dB scale) fade out */
    tail_db = target_db - tail_db;
    if (realsize == 4) {
        for (n = early_length; n < target_length; n++) {
            db = (double)(n - early_length) /
                (double)(target_length - early_length - 1) * tail_db;
            ((float *)imp)[n] = ((float *)imp)[n] * pow(10, db / 20);
        }
    } else {
        for (n = early_length; n < target_length; n++) {
            db = (double)(n - early_length) /
                (double)(target_length - early_length - 1) * tail_db;
            ((double *)imp)[n] = ((double *)imp)[n] * pow(10, db / 20);
        }
    }
    memset(&((uint8_t *)imp)[target_length * realsize], 0,
           (length - target_length) * realsize);
}

static void
make_natural_decay(void *imp,
                   int length,
                   int realsize,
                   int early_length,
                   int *new_length)
{
    double tail_db, db, init_db, slope, target_db, cur_db, base_att;
    int start, n, block_len;
    
    *new_length = length;
    
    if (length <= 4 * early_length) {
        elog(LOGDEBUG, "too short reverb");
        return;
    }

    /* calculate tail dB (max of 'early_length' samples) */
    tail_db = get_max_db(imp, length, length - early_length, early_length,
                         realsize);

    /* tail_db is considered to be the noise, now find a linear slope in the
       dB scale from start + early_length to where the dB reaches tail_db + 20*/
    init_db = get_max_db(imp, length, early_length, early_length, realsize);

    elog(LOGDEBUG, "init/tail volume: %f/%f dB", init_db, tail_db);
    
    if (init_db < tail_db + 15) {
        elog(LOGDEBUG, "too small volume difference between head and tail");
        return;
    }
    db = tail_db + 20;
    for (n = 0, start = 2 * early_length;
         start + early_length <= length && db > tail_db + 10;
         start += early_length, n++)
    {
        db = get_max_db(imp, length, start, early_length, realsize);
        /*elog(LOGDEBUG, "%d %f", n, db);*/
    }
    slope = (init_db - db) / n;
    elog(LOGDEBUG, "slope: %f dB per %d samples", slope, early_length);

    base_att = 1.0;
    block_len = early_length;
    for (target_db = db - slope;
         start < length;
         start += block_len, target_db -= slope)
    {
        /*elog(LOGDEBUG, "%d target_db %f, base_att %f",
             start, target_db, base_att);
        */
        if (start + block_len > length) {
            block_len = length - start;
        }
        if (target_db < -160) {
            *new_length = start;
            break;
        }
        if (realsize == 4) {
            for (n = start; n < start + block_len; n++) {
                ((float *)imp)[n] *= base_att;
            }
        } else {
            for (n = start; n < start + block_len; n++) {
                ((double *)imp)[n] *= base_att;
            }
        }
        cur_db = get_max_db(imp, length, start, block_len, realsize);
        /*elog(LOGDEBUG, "cur_db %f", cur_db);*/
        if (cur_db <= target_db) {
            continue;
        }
        base_att *= pow(10, (target_db - cur_db) / 20);
        if (realsize == 4) {
            for (n = start; n < start + block_len; n++) {
                db = (double)(n - start) / (double)block_len *
                    (target_db - cur_db);
                ((float *)imp)[n] *= pow(10, db / 20);
            }
        } else {
            for (n = start; n < start + block_len; n++) {
                db = (double)(n - start) / (double)block_len *
                    (target_db - cur_db);
                ((double *)imp)[n] *= pow(10, db / 20);
            }
        }
        /*elog(LOGDEBUG, "applied slope (%f)", target_db - cur_db);*/
    }
    if (*new_length < length) {
        elog(LOGDEBUG, "response shortened, original %d, new %d",
             length, *new_length);
        memset((uint8_t *)imp + realsize * (*new_length), 0,
               (length - (*new_length)) * realsize);
    }
}
                   

static void
fadein(void *imp,
       int length,
       int mute_length,
       int fadein_length,
       int realsize)
{
    double x;
    int n;
    
    elog(LOGDEBUG, "fadein_length: %d, mute_length: %d, length: %d",
         fadein_length, mute_length, length);
    if (fadein_length < 0 || mute_length < 0 || length < 0) {
        return;
    }
    if (length < mute_length) {
        memset(imp, 0, length * realsize);
        return;
    }
    if (fadein_length > mute_length) {
        fadein_length = mute_length;
    }
    memset(imp, 0, (mute_length - fadein_length) * realsize);

    /* fadein with function 0.5 + 0.5 * sin(x), -M_PI/2 <= x <= M_PI/2 */
    imp = &((uint8_t *)imp)[(mute_length - fadein_length) * realsize];
    if (realsize == 4) {
        for (n = 0; n < fadein_length; n++) {
            x = M_PI * ((double)(n + 1) / (double)fadein_length) - M_PI / 2;
            ((float *)imp)[n] = (0.5 + 0.5 * sin(x)) * ((float *)imp)[n];
        }
    } else {
        for (n = 0; n < fadein_length; n++) {
            x = M_PI * ((double)(n + 1) / (double)fadein_length) - M_PI / 2;
            ((double *)imp)[n] = (0.5 + 0.5 * sin(x)) * ((double *)imp)[n];
        }
    }
    /*
    {
        FILE *stream;
        stream = fopen("fadein_curve", "wt");
        for (n = 0; n < mute_length - fadein_length; n++) {
            fprintf(stream, "%.16e\n", 0.0);
        }
        for (n = 0; n < fadein_length; n++) {
            x = M_PI * ((real_t)(n + 1) / (real_t)fadein_length) - M_PI / 2;
            fprintf(stream, "%.16e\n", (0.5 + 0.5 * sin(x)) * 0.016);
        }
        fclose(stream);
    }
    */
}

static int
apply_response(struct bfirfilter *target,
               const void *source[],
               const int lengths[],               
               double _predelay,
               double _scale_length,
               bool_t decorrelate,
               bool_t apply_natural_decay,
               double fade_in,
               double attenuate,
               double early_attenuate,
               double *tail_volume,
               double *length,
               double *new_rt60,
               bool_t *truncated,
               int realsize,
               const char dump_filename[])
{
    int ri, scale_length, predelay, result_length, n, early_length;
    void *result, *result2, *result_arr[VCU_VALID_RATES_COUNT];
    int result_len_arr[VCU_VALID_RATES_COUNT];
    double scale, r, init_db;
    FILE *stream;

    elog(LOGDEBUG, "scale_length %f, predelay %f, decorr %d",
         _scale_length, _predelay, decorrelate);
    *length = 0.0;
    *tail_volume = 0.0;
    *new_rt60 = 0.0;
    *truncated = false;
    for (ri = 0; ri < target->n_rates; ri++) {
        /* allocate and copy to working copy of response */
        predelay = (int)(_predelay * target->rates[ri]);
        early_length = (int)(settings.early_length * (double)target->rates[ri]);
        scale_length = (int)(_scale_length * (double)(predelay + lengths[ri]));
        if ((result = alignalloc((predelay + lengths[ri]) *
                                 realsize)) == NULL)
        {
            elog(LOGDEBUG, "could not allocate %d bytes (predelay: %d; ri: %d; "
                 "lengths[ri]: %d; realsize: %d)",
                 (predelay + lengths[ri]) * realsize, predelay, ri,
                 lengths[ri], realsize);
            return VCU_ERR_NO_MEMORY;
        }
        memset(result, 0, predelay * realsize);
        memcpy(&((uint8_t *)result)[predelay * realsize], source[ri],
               lengths[ri] * realsize);

        if (predelay > 0) {
            /* create new direct sound */
            elog(LOGDEBUG, "adding new direct sound");
            if (realsize == 4) {
                ((float *)result)[0] = 1.0;
            } else {
                ((double *)result)[0] = 1.0;
            }
        }

        /* apply attenuation */
        if (attenuate != 0) {
            elog(LOGDEBUG, "attenuating %f dB", attenuate);
            scale = pow(10.0, attenuate / 20);
            if (realsize == 4) {
                for (n = 0; n < lengths[ri]; n++) {
                    ((float *)result)[predelay + n] *= scale;
                }
                ((float *)result)[0] *= scale;
            } else {
                for (n = 0; n < lengths[ri]; n++) {
                    ((double *)result)[predelay + n] *= scale;
                }
                ((double *)result)[0] *= scale;
            }
        }
        result_length = predelay + lengths[ri];
        
        /* apply natural decay */
        if (apply_natural_decay) {
            make_natural_decay(result, result_length, realsize, early_length,
                               &result_length);
        }
        
        /* apply decorrelate mode */
        if (decorrelate) {
            elog(LOGDEBUG, "decorrelating reverb start");
            result2 = decorr_reverb(result, result_length, target->rates[ri],
                                    0.0, true, realsize);
            elog(LOGDEBUG, "decorrelating reverb end");
            if (result2 == NULL) {
                free(result);
                elog(LOGDEBUG, "decorrelation failed");
                return VCU_ERR_NO_MEMORY;
            }
        }        

        /* apply scale length */
        if (scale_length > 0 && scale_length < result_length) {
            shorten_reverb(result, result_length, early_length, scale_length,
                           realsize);
            elog(LOGDEBUG, "shortened response (length %d -> %d)",
                 result_length, scale_length);
            result_length = scale_length;
        }
        init_db = get_max_db(result, result_length, 0, early_length, realsize);

        /* apply fadein */
        if (fade_in > 0.0) {
            elog(LOGDEBUG, "applying fade in %fs", fade_in);
            n = (int)(fade_in * (double)target->rates[ri]);
            fadein(result, result_length, n, n / 4, realsize);
        }

        /* apply early attenuation */
        if (early_attenuate > 0.0) {
            elog(LOGDEBUG, "applying early attenuation %fdB", early_attenuate);
            scale = pow(10, -early_attenuate / 20.0);
            if (realsize == 4) {
                for (n = 0; n < early_length && n < result_length; n++) {
                    ((float *)result)[n] *= scale;
                }
            } else {
                for (n = 0; n < early_length && n < result_length; n++) {
                    ((double *)result)[n] *= scale;
                }
            }
            for (n = early_length;
                 n < 3 * early_length && n < result_length;
                 n++)
            {
                r = (double)(n - early_length) / (double)(2 * early_length);
                scale = pow(10, -(1.0 - r) * early_attenuate / 20.0);
                if (realsize == 4) {
                    ((float *)result)[n] *= scale;
                } else {
                    ((double *)result)[n] *= scale;
                }
            }
        }
        
        /* truncate the response if necessary */
        if (result_length > target->f[ri].block_size * target->f[ri].n_blocks) {
            elog(LOGDEBUG, "truncating response from %d to %d",
                 result_length,
                 target->f[ri].block_size * target->f[ri].n_blocks);
            result_length = target->f[ri].block_size * target->f[ri].n_blocks;
            *truncated = true;
        }

        /* calculate parameters */
        elog(LOGDEBUG, "calculating parameters");        
        if (*length == 0 ||
            (double)result_length / (double)target->rates[ri] > *length)
        {
            *length = (double)result_length / (double)target->rates[ri];
        }
        *tail_volume += get_max_db(result, result_length,
                                   result_length - early_length,
                                   early_length, realsize);
        if (ri == 0 || *new_rt60 > 0) {
            n = calc_reverb_time
                (&((uint8_t *)result)[3 * early_length * realsize],
                 result_length - 3 * early_length,
                 predelay + 3 * early_length,
                 early_length, 60.0, init_db, realsize);
            if (n <= 0) {
                *new_rt60 = 0;
            } else {
                *new_rt60 += (double)n / (double)target->rates[ri];
            }
        }
            
        /* mix in dry signal */
        if (settings.dry_level > 0) {
            elog(LOGDEBUG, "mix in dry signal");
            if (realsize == 4) {
                ((float *)result)[0] += settings.dry_level;
            } else {
                ((double *)result)[0] += settings.dry_level;
            }
        }
        
        /* dump to file - for debugging */
        if (dump_filename != NULL) {
            elog(LOGDEBUG, "writing: \"%s\"...", dump_filename);
            stream = fopen(dump_filename, "wt");
            /*
            if (realsize == 4) {
                for (n = 0; n < result_length; n++) {
                    r = log10(((float *)result)[n]) * 20;
                    if (isnan(r) || isinf(r)) {
                        r = -200;
                    }
                    fprintf(stream, "%.16e\n", r);
                }
            } else {
                for (n = 0; n < result_length; n++) {
                    r = log10(((double *)result)[n]) * 20;
                    if (isnan(r) || isinf(r)) {
                        r = -200;
                    }
                    fprintf(stream, "%.16e\n", r);
                }
            }
            */
            if (realsize == 4) {
                for (n = 0; n < result_length; n++) {
                    fprintf(stream, "%.16e\n", ((float *)result)[n]);
                }
            } else {
                for (n = 0; n < result_length; n++) {
                    fprintf(stream, "%.16e\n", ((double *)result)[n]);
                }
            }
            fclose(stream);
            elog(LOGDEBUG, "done");
        }

        /* put the result to the bfir filter */
        elog(LOGDEBUG, "converting to bfir filter (result_length %d, ri %d)",
             result_length, ri);
        memset(result_arr, 0, sizeof(result_arr));
        result_arr[ri] = result;
        result_len_arr[ri] = result_length;
        bfirfilter_from_raw((const void **)result_arr,
                            result_len_arr,
                            target,
                            ri,
                            realsize);
        
        free(result);
    }

    *tail_volume /= target->n_rates;
    *new_rt60 /= target->n_rates;

    elog(LOGDEBUG, "tail volume: %f dB; rt60: %f s", *tail_volume, *new_rt60);
    
    return 0;
}

int
reverb_settings(double bformat_directivity_factor,
                double early_response_length,
                double direct_sound_length,
                double dry_level,
                int response_mode,
                int decorrelate_mode,
                bool_t force_delay_diff,
                bool_t force_decay,
                int source_layout,
                int n_stages,
                const double azimuth[],
                const double elevation[],
                const double width[],
                const double height[])
{
    static bool_t isinit = false;
    int n;

    if (!isinit) {
        isinit = true;
        memset(&settings, 0, sizeof(settings));
    }
    
    if (bformat_directivity_factor < REVERB_BFORMAT_DIRECTIVITY_MIN ||
        bformat_directivity_factor > REVERB_BFORMAT_DIRECTIVITY_MAX ||
        early_response_length < 0 ||
        direct_sound_length < 0 ||
        (response_mode != VCU_REVERB_RESP_MODE_FULL &&
         response_mode != VCU_REVERB_RESP_MODE_MUTE_DIRECT &&
         response_mode != VCU_REVERB_RESP_MODE_MUTE_EARLY &&
         response_mode != VCU_REVERB_RESP_MODE_STAGE_ADAPTED) ||
        (decorrelate_mode != VCU_REVERB_DECORR_MODE_ALL &&
         decorrelate_mode != VCU_REVERB_DECORR_MODE_NEEDED &&
         decorrelate_mode != VCU_REVERB_DECORR_MODE_NONE) ||
        (source_layout != VCU_REVERB_SOURCE_LAYOUT_CENTER &&
         source_layout != VCU_REVERB_SOURCE_LAYOUT_STEREO &&
         source_layout != VCU_REVERB_SOURCE_LAYOUT_LRC))
    {
        return VCU_ERR_INVALID_VALUE;
    }
    if (n_stages > 0) {
        if (azimuth == NULL || elevation == NULL || width == NULL ||
            height == NULL)
        {
            return VCU_ERR_INVALID_VALUE;
        }
        settings.stage_count = n_stages;
        free(settings.stage);
        settings.stage = malloc(n_stages * sizeof(struct stage));
        for (n = 0; n < n_stages; n++) {
            settings.stage[n].azimuth = azimuth[n];
            settings.stage[n].elevation = elevation[n];
            settings.stage[n].width = width[n];
            settings.stage[n].height = height[n];
        }
    } else {
        settings.stage_count = 0;
        free(settings.stage);
        settings.stage = NULL;
    }
    settings.bformat_directivity = bformat_directivity_factor;
    settings.early_length = early_response_length;
    settings.direct_length = direct_sound_length;
    settings.dry_level = dry_level;
    settings.force_delay_diff = !!force_delay_diff;
    settings.force_decay = !!force_decay;
    settings.response_mode = response_mode;
    settings.decorrelate_mode = decorrelate_mode;
    switch (source_layout) {
    case VCU_REVERB_SOURCE_LAYOUT_CENTER:
        settings.n_sources = 1;
        break;
    case VCU_REVERB_SOURCE_LAYOUT_STEREO:
        settings.n_sources = 2;
        break;
    case VCU_REVERB_SOURCE_LAYOUT_LRC:
        settings.n_sources = 3;
        break;
    }
    return 0;
}

double
reverb_postprocess_maxlength(double original_length,
                             double predelay)
{
    return original_length + predelay;
}

int
reverb_apply_balance(const double xyz_balance[3],
                     int n_reverb_speakers,
                     const double azimuth[],
                     const double elevation[],
                     double volume[])
{
    double pos, neg;
    int n;

    if (xyz_balance == NULL || n_reverb_speakers < 1 || azimuth == NULL ||
        elevation == NULL || volume == NULL)
    {
        return VCU_ERR_INVALID_VALUE;
    }

    memset(volume, 0, n_reverb_speakers * sizeof(double));

    /* X balance (front-back) */
    for (n = 0, pos = neg = 0.0; n < n_reverb_speakers; n++) {
        /* we use 80 degrees as limit, since speakers playing at 90 degrees
           sounds as coming from back (or is it just me?) */
        if (azimuth[n] < 80 && azimuth[n] > -80) {
            pos += 1.0;
        } else {
            neg += 1.0;
        }
    }
    if (pos != 0.0 && neg != 0.0) {
        for (n = 0; n < n_reverb_speakers; n++) {
            if (azimuth[n] < 80 && azimuth[n] > -80) {
                volume[n] += (xyz_balance[0] * neg / (pos + neg));
            } else {
                volume[n] += -(xyz_balance[0] * pos / (pos + neg));
            }
        }
    }

    /* Y balance (left-right) */
    for (n = 0, pos = neg = 0.0; n < n_reverb_speakers; n++) {
        if (azimuth[n] > 0) {
            pos += 1.0;
        } else if (azimuth[n] < 0) {
            neg += 1.0;
        }
    }
    if (pos != 0.0 && neg != 0.0) {
        for (n = 0; n < n_reverb_speakers; n++) {
            if (azimuth[n] > 0) {
                volume[n] += (xyz_balance[1] * neg / (pos + neg));
            } else if (azimuth[n] < 0) {
                volume[n] += -(xyz_balance[1] * pos / (pos + neg));
            }
        }
    }

    /* Z balance (up-down) */    
    for (n = 0, pos = neg = 0.0; n < n_reverb_speakers; n++) {
        if (elevation[n] > 0) {
            pos += 1.0;
        } else {
            neg += 1.0;
        }
    }    
    if (pos != 0.0 && neg != 0.0) {
        for (n = 0; n < n_reverb_speakers; n++) {
            if (elevation[n] > 0) {
                volume[n] += (xyz_balance[2] * neg / (pos + neg));
            } else {
                volume[n] += -(xyz_balance[2] * pos / (pos + neg));
            }
        }
    }
    return 0;
}

bool_t
reverb_test_receiver_directions(const double azimuth[],
                                const double elevation[],
                                int n_receivers)
{
    int n, i;
    
    /*

      Rules for a traditional reverb program, as seen per set:

        - if there is only one impulse response (pair, if stereo source), the
          azimuth/elevation must be 0/0
        - if there are more than one, they must either be symmetrically placed
          around the X axis, or only be to the left side.
        - it is not allowed to have more than one impulse response (pair) in
          the same direction.
        - any with elevation 90 or -90 must have azimuth 0
      
    */

    if (azimuth == NULL || elevation == NULL || n_receivers < 1) {
        elog(LOGDEBUG, "invalid parameter azimuth %p elevation %p "
             "n_receivers %d", azimuth, elevation, n_receivers);
        return false;
    }

    if (n_receivers == 1 && (azimuth[0] != 0.0 || elevation[0] != 0.0)) {
        elog(LOGDEBUG, "one receiver, but not in 0/0 direction");
        return false;
    }

    for (n = 0; n < n_receivers; n++) {
        if ((elevation[n] == -90.0 || elevation[n] == 90.0) &&
            azimuth[n] != 0.0)
        {
            elog(LOGDEBUG, "elevation is +/-90 but azimuth not 0");
            return false;
        }
        if (azimuth[n] > 180.0 || azimuth[n] < -180.0 ||
            elevation[n] > 90.0 || elevation[n] < -90.0)
        {
            elog(LOGDEBUG, "azimuth or elevation is out of range");
            return false;
        }
        for (i = 0; i < n_receivers; i++) {
            if (i != n && azimuth[n] == azimuth[i] &&
                elevation[n] == elevation[i])
            {
                elog(LOGDEBUG, "at least two directions are same");
                return false;
            }
        }
    }
    for (n = 0; n < n_receivers; n++) {
        if (azimuth[n] < 0.0) {
            break;
        }
    }
    if (n == n_receivers) {
        /* one-sided */
        return true;
    }
    for (n = 0; n < n_receivers; n++) {
        if (azimuth[n] == 0.0 || azimuth[n] == -0.0 ||
            azimuth[n] == 180.0 || azimuth[n] == -180.0)
        {
            continue;
        }
        for (i = 0; i < n_receivers; i++) {
            if (i != n && azimuth[n] == -azimuth[i] &&
                elevation[n] == elevation[i])
            {
                break;
            }
        }
        if (i == n_receivers) {
            elog(LOGDEBUG, "receivers are neither one-sided or symmetric");
            return false;
        }
    }    
    
    return true;
}

int
reverb_preprocess(const struct filterprg *fprg,
                  int realsize,
                  double target_length,
                  int n_reverb_speakers,
                  const double azimuth[],
                  const double elevation[],
                  struct bfirfilter target[],
                  double *orig_rt60,
                  double *orig_tail_volume,
                  double *new_rt60,
                  double *new_tail_volume,
                  double *new_length,
                  bool_t *truncated,
                  void (*progress_callback)(int channel_index,
                                            int n_channels,
                                            int from_rate,
                                            int to_rate,
                                            bool_t deriving))
{
    int buflengths[VCU_VALID_RATES_COUNT], load_index, load_count, prepf_count;
    double tail_volume, length, rt60, fade_in, early_attenuate, diffdelay, r;
    int unused_minpos, *revsort, n_sources, *channel_list, *speakers;
    void *buf[VCU_VALID_RATES_COUNT], *dest[VCU_VALID_RATES_COUNT];
    int n, i, j, k, a, b, minpos, err, *receiver, *set, *used_set;
    int post_load_source_count, destlen[VCU_VALID_RATES_COUNT];
    bool_t decorrelate, do_decorrelate, *used, _truncated;
    float df, xf, yf, zf, iwf, ixf, iyf, izf, iwmulf;
    double fade_in_length, early_attenuate_length;
    double d, x, y, z, iw, ix, iy, iz, iwmul;
    double elevation_correction;
    uint32_t random_state[3];
    char dump_filename[200];
    struct prepfilter prepf;
    struct direction *dir;

    /*

      The task is to assign a unique set (1 - 3 depending on source layout,
      settings.n_sources) of impulse responses to each reverb speaker.

      1. The closest impulse response. This is made the same way for both
         traditional and b-format reverb formats.
         
         - reverb speakers are first sorted after closest to vector 0/0, meaning
           that frontal speakers will get priority in finding the best suitable
           impulse response
         - least difference in angle
         - if only responses to one side (must be left), mirror the directions
           when comparing
         - for side and back speakers, where fabs(azimuth) >= 80, we rather
           assign a back impulse response (meaning fabs(azimuth) >= 90) than a
           front one, if available
         - if not very large angle differences, rather try to use all impulse
           responses than always picking the closest one.
         - if the impulse response has been used already, pick it from another
           set if available

      2. Source layout mismatch compensation. In some cases, the source layout
         in the reverb program is not the same as in the reverb engine.

           Source count     How to compensate
         engine / program
            1   /   3       Only the center source is loaded.
            1   /   2       Left and right source is mixed into a fake center.
            2   /   3       Only the left and right source are loaded.
            2   /   1       Center = fake left, center decorr = fake right +
                            delay difference
            3   /   1       Center = fake left, center decorr = center, center
                            decorr = fake right, delay difference
            3   /   2       Mix left right source and decorrelate for a fake
                            center.

      3. Apply reverb engine parameters

         - decorrelation
         - early muting
         - etc

         
      4. Attenuation compensation. This is only made for the traditional reverb
         format.

         - Azimuth and elevation is handled separately. If there is large
           difference between speaker and impulse response direction,
           compensation is possibly applied through early response attenuation.

     */

    elog(LOGDEBUG, "preprocess reverb program start");

    elog(LOGDEBUG, "sanity check input parameters");
    if (fprg == NULL || (fprg->type != FILTERPRG_TYPE_REVERB_TRADITIONAL &&
                         fprg->type != FILTERPRG_TYPE_REVERB_BFORMAT) ||
        n_reverb_speakers < 1 || azimuth == NULL ||
	elevation == NULL || target == NULL || truncated == NULL ||
        orig_rt60 == NULL || orig_tail_volume == NULL ||
        new_length == NULL || new_tail_volume == NULL || new_rt60 == NULL ||
        (realsize != sizeof(float) && realsize != sizeof(double)))
    {
        elog(LOGDEBUG, "fprg %p; n_reverb_speakers %d; azimuth %p; "
            "elevation %p; target %p; new_length %p; new_tail_volume %p; "
            "realsize %d", fprg, n_reverb_speakers, azimuth, elevation,
             target, new_length, new_tail_volume, realsize);
	return VCU_ERR_INVALID_VALUE;
    }
    *orig_tail_volume = 0.0;
    *orig_rt60 = 0.0;
    *new_length = 0.0;
    *new_tail_volume = 0.0;
    *new_rt60 = 0.0;
    *truncated = false;
    n_sources = fprg->n_channels / fprg->n_sets / fprg->n_receivers;
    if (fprg->type == FILTERPRG_TYPE_REVERB_BFORMAT) {
        n_sources /= 4;
    }
    if (n_sources < 1 || n_sources > 3) {
        elog(LOGDEBUG, "invalid number of sources: %d", n_sources);
        return VCU_ERR_INVALID_VALUE;
    }
    
    channel_list = alloca(4 * n_sources * sizeof(int));
    receiver = alloca(n_reverb_speakers * sizeof(int));
    set = alloca(n_reverb_speakers * sizeof(int));
    dir = alloca(n_reverb_speakers * sizeof(struct direction));
    used = alloca(fprg->n_receivers * sizeof(bool_t));
    used_set = alloca(fprg->n_sets * sizeof(int));
    revsort = alloca(n_reverb_speakers * sizeof(int));
    memset(buf, 0, sizeof(buf));

    /* sort speakers, after closest to 0/0*/
    for (n = 0; n < n_reverb_speakers; n++) {
        dir[n].index = n;
        dir[n].az = azimuth[n];
        dir[n].el = elevation[n];
    }
    qsort(dir, n_reverb_speakers, sizeof(struct direction), anglecmp);
    
    /* find best suiting receiver for each speaker */
    elog(LOGDEBUG, "find best suiting receiver for each speaker");
    for (n = 0; n < n_reverb_speakers; n++) {
        minpos = bestreceiver(dir[n].az, dir[n].el, fprg->recv_azimuth,
                              fprg->recv_elevation, NULL, fprg->n_receivers);
        elog(LOGDEBUG, "best suited for a%.0f e%.0f is receiver %d "
             "(a%.0f e%.0f)", dir[n].az, dir[n].el,
             minpos, fprg->recv_azimuth[minpos], fprg->recv_elevation[minpos]);

        /* if the impulse response already has been used, select from
           another set if available */
        set[dir[n].index] = 0;
        do {
            for (i = 0; i < n; i++) {
                if (receiver[dir[i].index] == minpos &&
                    set[dir[i].index] == set[dir[n].index])
                {
                    set[dir[n].index]++;
                    break;
                }
            }
        } while (i != n && set[dir[n].index] < fprg->n_sets);
        if (set[dir[n].index] == fprg->n_sets) {
            /* find best receiver from unused impulse responses, if available */
            memset(used, 0, fprg->n_receivers * sizeof(bool_t));
            for (j = 0; j < fprg->n_receivers; j++) {
                memset(used_set, -1, sizeof(fprg->n_sets * sizeof(int)));
                for (i = 0; i < n && !used[j]; i++) {
                    if (receiver[dir[i].index] == j) {
                        for (k = 0; used_set[k] != -1; k++) {
                            if (used_set[k] == set[dir[i].index]) {
                                break;
                            }
                        }
                        if (used_set[k] == -1) {
                            used_set[k] = set[dir[i].index];
                            if (k == fprg->n_sets - 1) {
                                elog(LOGDEBUG, "receiver %d is used", j);
                                used[j] = true;
                            }
                        }
                    }
                }
            }
            unused_minpos = bestreceiver(dir[n].az, dir[n].el,
                                         fprg->recv_azimuth,
                                         fprg->recv_elevation, used,
                                         fprg->n_receivers);
            elog(LOGDEBUG, "unused_minpos: %d", unused_minpos);
            if (unused_minpos != -1) {
                /* only use unused_minpos, if the match is not too bad */
                if ((isback(fprg->recv_azimuth[unused_minpos], 200) &&
                     isback(dir[n].az, 200)) ||
                    (!isback(fprg->recv_azimuth[unused_minpos], 160) &&
                     !isback(dir[n].az, 160)))
                {
                    minpos = unused_minpos;
                }
            }
            set[dir[n].index] = 0;
            do {
                for (i = 0; i < n; i++) {
                    if (receiver[dir[i].index] == minpos &&
                        set[dir[i].index] == set[dir[n].index])
                    {
                        set[dir[n].index]++;
                        break;
                    }
                }
            } while (i != n && set[dir[n].index] < fprg->n_sets);
            if (set[dir[n].index] == fprg->n_sets) {
                set[dir[n].index] = 0;
            }
        }
        
        receiver[dir[n].index] = minpos;
        elog(LOGDEBUG, "selecting receiver %d in set %d for speaker %d",
             receiver[dir[n].index], set[dir[n].index], dir[n].index);
    }

    /* order reverb speakers so impulse responses only need to be loaded
       once */
    speakers = alloca(n_reverb_speakers * sizeof(int));
    load_count = 0;
    for (n = 0, j = 0; n < n_reverb_speakers; n++) {
        for (i = 0; i < j; i++) {
            if (speakers[i] == n && set[speakers[i]] == set[n]) {
                break;
            }
        }
        if (j != i) {
            continue;
        }
        load_count++;
        for (i = 0; i < n_reverb_speakers; i++) {
            if (receiver[n] == receiver[i] && set[n] == set[i]) {
                speakers[j++] = i;
            }
        }
    }
    
    /* apply the impulse responses */
    elog(LOGDEBUG, "apply impulse responses (decorrelate mode: %d)",
         settings.decorrelate_mode);
    post_load_source_count = -1;
    load_index = 0;
    prepf_count = 0;
    memset(&prepf, 0, sizeof(prepf));
    for (j = 0; j < n_reverb_speakers; j++) {
        n = speakers[j];
        elog(LOGDEBUG, "j: %d; n: %d; receiver: %d", j, n, receiver[n]);
        if (j == 0 || receiver[speakers[j]] != receiver[speakers[j-1]]) {
            if (j > 0) {
                filterprg_free_prepfilter_contents(&prepf);
                memset(&prepf, 0, sizeof(prepf));
            }
            elog(LOGDEBUG, "rev speaker %d: loading new prepfilter (%d)",
                 n, receiver[n]);
            switch (fprg->type) {
            case FILTERPRG_TYPE_REVERB_TRADITIONAL:
                if (settings.n_sources == 1 && n_sources == 3) {
                    elog(LOGDEBUG, "only load center source");
                    channel_list[0] =
                        receiver[n] +
                        2 * fprg->n_receivers +
                        set[n] * n_sources * fprg->n_receivers;
                    post_load_source_count = 1;
                } else if (settings.n_sources == 2 && n_sources == 3) {
                    elog(LOGDEBUG, "only load left and right source");
                    for (a = 0; a < 2; a++) {
                        channel_list[a] =
                            receiver[n] +
                            a * fprg->n_receivers +
                            set[n] * n_sources * fprg->n_receivers;
                    }
                    post_load_source_count = 2;
                } else {
                    /* default case, load all sources in reverb program */
                    for (a = 0; a < n_sources; a++) {
                        channel_list[a] =
                            receiver[n] +
                            a * fprg->n_receivers +
                            set[n] * n_sources * fprg->n_receivers;
                    }
                    post_load_source_count = n_sources;
                }
                elog(LOGDEBUG, "loading %d channels", n_sources);
                if ((err = filterprg_load(fprg, &prepf, NULL,
                                          channel_list, 0,
                                          post_load_source_count,
                                          target[0].rates,
                                          target[0].n_rates,
                                          realsize, progress_callback,
                                          load_index,
                                          post_load_source_count * load_count))
                    != 0)
                {
                    goto reverb_preprocess_abort;
                }
                *orig_rt60 +=
                    calc_reverb_time_prepf(&prepf, fprg->predelay, 60.0);
                *orig_tail_volume += calc_tail_volume_prepf(&prepf);
                prepf_count += 1;
                load_index += post_load_source_count;
                if (settings.n_sources == 1 && n_sources == 2) {
                    elog(LOGDEBUG, "mixing left/right into a fake center");
                    for (i = 0; i < target[0].n_rates; i++) {
                        if (realsize == 4) {
                            for (k = 0; k < prepf.lengths[i]; k++) {
                                ((float *)prepf.filters[0][i])[k] +=
                                    ((float *)prepf.filters[1][i])[k];
                            }
                        } else {
                            for (k = 0; k < prepf.lengths[i]; k++) {
                                ((double *)prepf.filters[0][i])[k] +=
                                    ((double *)prepf.filters[1][i])[k];
                            }
                        }
                    }
                } else if (settings.n_sources == 3 && n_sources == 2) {
                    elog(LOGDEBUG, "mixing left/right into a new fake center");
                    prepf.n_channels = 3;
                    for (i = 0; i < target[0].n_rates; i++) {
                        if ((prepf.filters[2][i] =
                             alignalloc(realsize * prepf.lengths[i])) == NULL)
                        {
                            err = VCU_ERR_NO_MEMORY;
                            goto reverb_preprocess_abort;
                        }
                        if (realsize == 4) {
                            for (k = 0; k < prepf.lengths[i]; k++) {
                                ((float *)prepf.filters[2][i])[k] =
                                    ((float *)prepf.filters[0][i])[k] +
                                    ((float *)prepf.filters[1][i])[k];
                            }
                        } else {
                            for (k = 0; k < prepf.lengths[i]; k++) {
                                ((double *)prepf.filters[2][i])[k] =
                                    ((double *)prepf.filters[0][i])[k] +
                                    ((double *)prepf.filters[1][i])[k];
                            }
                        }
                        decorr_reverb(prepf.filters[2][i],
                                      prepf.lengths[i],
                                      target[0].rates[i],
                                      0.0, true, realsize);
                    }
                }
                break;
            case FILTERPRG_TYPE_REVERB_BFORMAT:

                if (settings.n_sources == 1 && n_sources == 3) {
                    elog(LOGDEBUG, "only load center source");
                    for (i = 0; i < 4; i++) {
                        channel_list[i] =
                            i +
                            receiver[n] * 4 +
                            2 * fprg->n_receivers * 4 +
                            set[n] * n_sources * fprg->n_receivers * 4;
                    }
                    post_load_source_count = 1;
                } else if (settings.n_sources == 2 && n_sources == 3) {
                    elog(LOGDEBUG, "only load left and right source");
                    for (a = 0; a < 2; a++) {
                        for (i = 0; i < 4; i++) {
                            channel_list[4*a+i] =
                                i +
                                receiver[n] * 4 +
                                a * fprg->n_receivers * 4 +
                                set[n] * n_sources * fprg->n_receivers * 4;
                        }
                    }
                    post_load_source_count = 2;
                } else {
                    /* default case, load all sources in reverb program */
                    for (a = 0; a < n_sources; a++) {
                        for (i = 0; i < 4; i++) {
                            channel_list[4*a+i] =
                                i +
                                /* receiver offset */
                                receiver[n] * 4 +
                                /* source offset */
                                a * fprg->n_receivers * 4 +
                                /* set offset */
                                set[n] * n_sources * fprg->n_receivers * 4;
                        }
                    }
                    post_load_source_count = n_sources;
                }
                elog(LOGDEBUG, "loading %d channels",
                     post_load_source_count * 4);
                if ((err = filterprg_load(fprg, &prepf, NULL,
                                          channel_list, 0,
                                          post_load_source_count * 4,
                                          target[0].rates,
                                          target[0].n_rates,
                                          realsize, progress_callback,
                                          load_index,
                                          post_load_source_count * 4 *
                                          load_count))
                    != 0)
                {
                    goto reverb_preprocess_abort;
                }
                *orig_rt60 +=
                    calc_reverb_time_prepf(&prepf, fprg->predelay, 60.0);
                *orig_tail_volume += calc_tail_volume_prepf(&prepf);
                prepf_count += 1;
                load_index += post_load_source_count * 4;
                elog(LOGDEBUG, "finished loading %d channels",
                     post_load_source_count * 4);
                
                if (settings.n_sources == 1 && n_sources == 2) {
                    elog(LOGDEBUG, "mixing left/right into a fake center");
                    for (i = 0; i < target[0].n_rates; i++) {
                        if (realsize == 4) {
                            for (k = 0; k < prepf.lengths[i]; k++) {
                                ((float *)prepf.filters[0][i])[k] +=
                                    ((float *)prepf.filters[4][i])[k];
                                ((float *)prepf.filters[1][i])[k] +=
                                    ((float *)prepf.filters[5][i])[k];
                                ((float *)prepf.filters[2][i])[k] +=
                                    ((float *)prepf.filters[6][i])[k];
                                ((float *)prepf.filters[3][i])[k] +=
                                    ((float *)prepf.filters[7][i])[k];
                            }
                        } else {
                            for (k = 0; k < prepf.lengths[i]; k++) {
                                ((double *)prepf.filters[0][i])[k] +=
                                    ((double *)prepf.filters[4][i])[k];
                                ((double *)prepf.filters[1][i])[k] +=
                                    ((double *)prepf.filters[5][i])[k];
                                ((double *)prepf.filters[2][i])[k] +=
                                    ((double *)prepf.filters[6][i])[k];
                                ((double *)prepf.filters[3][i])[k] +=
                                    ((double *)prepf.filters[7][i])[k];
                            }
                        }
                    }
                } else if (settings.n_sources == 3 && n_sources == 2) {
                    elog(LOGDEBUG, "mixing left/right into a new fake center");
                    decorr_get_random_state(random_state);
                    prepf.n_channels = 12;
                    for (i = 0; i < target[0].n_rates; i++) {
                        if ((prepf.filters[8][i] =
                             alignalloc(realsize * prepf.lengths[i])) == NULL ||
                            (prepf.filters[9][i] =
                             alignalloc(realsize * prepf.lengths[i])) == NULL ||
                            (prepf.filters[10][i] =
                             alignalloc(realsize * prepf.lengths[i])) == NULL ||
                            (prepf.filters[11][i] =
                             alignalloc(realsize * prepf.lengths[i])) == NULL)
                        {
                            err = VCU_ERR_NO_MEMORY;
                            goto reverb_preprocess_abort;
                        }
                        if (realsize == 4) {
                            for (k = 0; k < prepf.lengths[i]; k++) {
                                ((float *)prepf.filters[8][i])[k] =
                                    ((float *)prepf.filters[0][i])[k] +
                                    ((float *)prepf.filters[4][i])[k];
                                ((float *)prepf.filters[9][i])[k] =
                                    ((float *)prepf.filters[1][i])[k] +
                                    ((float *)prepf.filters[5][i])[k];
                                ((float *)prepf.filters[10][i])[k] =
                                    ((float *)prepf.filters[2][i])[k] +
                                    ((float *)prepf.filters[6][i])[k];
                                ((float *)prepf.filters[11][i])[k] =
                                    ((float *)prepf.filters[3][i])[k] +
                                    ((float *)prepf.filters[7][i])[k];
                            }
                        } else {
                            for (k = 0; k < prepf.lengths[i]; k++) {
                                ((double *)prepf.filters[8][i])[k] =
                                    ((double *)prepf.filters[0][i])[k] +
                                    ((double *)prepf.filters[4][i])[k];
                                ((double *)prepf.filters[9][i])[k] =
                                    ((double *)prepf.filters[1][i])[k] +
                                    ((double *)prepf.filters[5][i])[k];
                                ((double *)prepf.filters[10][i])[k] =
                                    ((double *)prepf.filters[2][i])[k] +
                                    ((double *)prepf.filters[6][i])[k];
                                ((double *)prepf.filters[11][i])[k] =
                                    ((double *)prepf.filters[3][i])[k] +
                                    ((double *)prepf.filters[7][i])[k];
                            }
                        }
                        for (k = 0; k < 4; k++) {
                            /* make sure same random state is used for
                               each decorrelation, so W X Y Z is decorrelated
                               in the same way */
                            decorr_set_random_state(random_state);
                            decorr_reverb(prepf.filters[8+k][i],
                                          prepf.lengths[i],
                                          target[0].rates[i],
                                          0.0, true, realsize);
                        }
                    }
                }
                break;
            default:
                /* should not happen */
                elog(LOGERROR, "unexpected program type %d", fprg->type);
                return VCU_ERR_INVALID_VALUE;
            }
        } else {
            elog(LOGDEBUG, "rev speaker %d: using previous prepfilter (%d)",
                 n, receiver[n]);
        }

        /* see if we should decorrelate */
        decorrelate = false;
        switch (settings.decorrelate_mode) {
        case VCU_REVERB_DECORR_MODE_ALL:
            decorrelate = true;
            elog(LOGDEBUG, "decorrelate, because decorr_mode_all");
            break;
        case VCU_REVERB_DECORR_MODE_NEEDED:
            decorrelate = false;
            if (j > 0 &&
                receiver[speakers[j]] == receiver[speakers[j-1]] &&
                set[speakers[j]] == set[speakers[j-1]])
            {
                switch (fprg->type) {
                case FILTERPRG_TYPE_REVERB_TRADITIONAL:
                    elog(LOGDEBUG, "impulse already used, decorrelate "
                         "(mode_needed)");
                    decorrelate = true;
                    break;
                case FILTERPRG_TYPE_REVERB_BFORMAT:
                    for (i = j - 1;
                         i >= 0 &&
                             receiver[speakers[j]] == receiver[speakers[i]] &&
                             set[speakers[j]] == set[speakers[i]];
                         i--)
                    {
                        if (anglediff(azimuth[speakers[i]],
                                      elevation[speakers[i]],
                                      azimuth[speakers[j]],
                                      elevation[speakers[j]]) <= 10.0)
                        {
                            decorrelate = true;
                            break;
                        }
                    }
                    if (decorrelate) {
                        elog(LOGDEBUG, "impulse already used with similar/same "
                             "direction, decorrelate (mode_needed)");
                    } else {
                        elog(LOGDEBUG, "impulse already used but with "
                             "different direction, do not decorrelate "
                             "(mode_needed)");
                    }
                    break;
                }
                break;
            } else {
                elog(LOGDEBUG, "impulse first use, do not decorrelate "
                     "(mode_needed)");
            }
            break;
        case VCU_REVERB_DECORR_MODE_NONE:
            elog(LOGDEBUG, "do not decorrelate, because decorr_mode_none");
            decorrelate = false;
            break;
        }
        
        /* should we fade in the response? */
        fade_in = 0.0;
        switch (settings.response_mode) {
        case VCU_REVERB_RESP_MODE_FULL:
            fade_in = 0.0;
            elog(LOGDEBUG, "mode_full: fade_in: %g", fade_in);
            break;
        case VCU_REVERB_RESP_MODE_MUTE_DIRECT:
            fade_in = settings.direct_length;
            elog(LOGDEBUG, "mode_mute_direct: fade_in: %g", fade_in);
            break;
        case VCU_REVERB_RESP_MODE_MUTE_EARLY:
            fade_in = settings.early_length;
            elog(LOGDEBUG, "mode_mute_early: fade_in: %g", fade_in);
            break;
        case VCU_REVERB_RESP_MODE_STAGE_ADAPTED:
            if (within_stage(azimuth[n], elevation[n])) {
                fade_in = settings.early_length;
                elog(LOGDEBUG, "mode_stage_adapted, within stage: fade_in: %g",
                     fade_in);
            } else {
                fade_in = 0.0;
                elog(LOGDEBUG, "mode_stage_adapted, outside stage: fade_in: %g",
                     fade_in);
            }
            break;
        }

        /*
          Calculate compensation parameters, as if we only have one
          center source impulse response in the reverb program. Whether to use
          these parameters or not are decided further down.
          
          There are two parameters. Difference delay between left and right,
          meaning that reflections should arrive earlier to a left impulse
          response than to a right, if the source is on the left side, and
          the other way around. Early attenuation lowers the reflection energy
          in the early part of the response.
          
          Difference delay need only to be used if reverb engine is stereo
          or LRC, and the reverb program is center.
          
          Early attenuation compensation is not used for B-format reverb
          programs, since B-format is full 3D, however the difference delay
          may be used.
        */
        elevation_correction = 0.0;
        if (azimuth[n] <= 45 && azimuth[n] >= -45) {
            /* front 90 degree slice, the more to the sides, the
               larger delay difference between left and right source.
               Range is 0 - 5 ms. No early attenuation. */
            r = (azimuth[n] < 0) ? -azimuth[n] : azimuth[n];
            r = r / 45;
            diffdelay = 0.005 * r;
            early_attenuate = 0.0;
            elog(LOGDEBUG, "front 90 slice: dd: %g, ea: %g",
                 diffdelay, early_attenuate);
        } else if (azimuth[n] <= 90 && azimuth[n] >= -90) {
            /* 90 - 45 degree slice both sides, the more to the sides,
               the larger early attenuation. Range is 0 - 5.0 ms.
               Left/right delay difference fixed to 5 ms. */
            r = (azimuth[n] < 0) ? -azimuth[n] : azimuth[n];
            r = (r - 45) / 45;
            diffdelay = 0.005;
            early_attenuate = 5.0 * r;
            elog(LOGDEBUG, "side 90 - 45 slice: dd: %g, ea: %g",
                 diffdelay, early_attenuate);
        } else if (azimuth[n] <= 135 && azimuth[n] >= -135) {
            /* 135 - 90 degree slice both sides, the more to the back,
               the larger early attenuation, and the less delay
               difference between left and right source.
               Range is 5.0 - 2.5 ms for left/right delay difference,
               and 5.0 - 10.0 ms for early attenuation. */
            r = (azimuth[n] < 0) ? -azimuth[n] : azimuth[n];
            r = (r - 90) / 45;
            diffdelay = 0.005 - 0.0025 * r;
            early_attenuate = 5.0 + 5.0 * r;
            elog(LOGDEBUG, "side 135 - 90 slice: dd: %g, ea: %g",
                 diffdelay, early_attenuate);
        } else {
            /* rear 90 degree slice. Early attenuation fixed to 10 ms,
               the more to the back, the less left/right delay
               differenc. Range is 2.5 - 0.0 ms. */
            r = (azimuth[n] < 0) ? -azimuth[n] : azimuth[n];
            r = (r - 135) / 45;
            diffdelay = 0.0025 - 0.0025 * r;
            early_attenuate = 10.0;
            elog(LOGDEBUG, "rear 90 slice: dd: %g, ea: %g",
                 diffdelay, early_attenuate);
        }
        if (elevation[n] > 45) {
            /* if elevation is larger than 45 degrees, left/right
               delay difference is reduced the closer to straight up
               the receiver is. Range is -0.0 - -5.0 ms. Delay
               difference is truncated to zero if necessary. */
            r = (elevation[n] - 45) / 45;
            diffdelay -= 0.005 * r;
            if (diffdelay < 0.0) {
                diffdelay = 0.0;
            }
            /* if reverb program impulse response is not much elevated,
               add some early attenuation, more the larger the
               receiver elevation. Range is 0.0 - 3.0 ms */
            elevation_correction = 3.0 * r;
            elog(LOGDEBUG, "elevation > 45: dd: %g, ec: %g",
                 diffdelay, elevation_correction);
        }
        if (elevation[n] < -45) {
            /* if elevation is less than -45 degrees, left/right delay
               difference is reduced. */
            r = -(elevation[n] - 45) / 45;
            diffdelay -= 0.005 * r;
            if (diffdelay < 0.0) {
                diffdelay = 0.0;
            }
            /* if reverb program impulse response is not much elevated,
               add some early attenuation. Range is 0.0 - 6.0 ms. Larger
               early attenuation range is used here than for positive
               elevations, because it is expected to come more reverb energy
               from above than from the ground */
            elevation_correction = 6.0 * r;
            elog(LOGDEBUG, "elevation < -45: dd: %g, ec: %g",
                 diffdelay, elevation_correction);
        }
        
        for (a = 0; a < settings.n_sources; a++) {

            /* make copies of parameters we may change here */
            early_attenuate_length = early_attenuate;
            fade_in_length = fade_in;
            do_decorrelate = decorrelate;
            
            /*
              If there is more than one reverb engine source, and there is
              either only one reverb program source, or there is some early
              attenuation and delay difference is forced, we test if we should
              apply any left/right delay difference.
             */
            if (settings.n_sources > 1 &&
                (n_sources == 1 ||
                 (settings.force_delay_diff && fade_in > 0.0)))
            {
                /*
                  If the current reverb engine source is center, or the current
                  speaker azimuth is on the same side as the reverb engine
                  source, don't apply the diffdelay.
                */
                if (a == 2 ||
                    (a == 0 && azimuth[n] >= 0) ||
                    (a == 1 && azimuth[n] <= 0))
                {
                    elog(LOGDEBUG, "diffdelay not applied: "
                         "reveng src idx %d; azimuth %g", a, azimuth[n]);
                } else {
                    fade_in_length += diffdelay;
                    elog(LOGDEBUG, "applying diffdelay: %gs", diffdelay);
                }
            } else {
                elog(LOGDEBUG, "diffdelay not applied: reveng sources %d; "
                    "revprg sources %d; force_delay_diff %d; fade_in %g",
                     settings.n_sources, n_sources, settings.force_delay_diff,
                     fade_in_length);
            }

            /* Do we need to decorrelate due to missing sources in reverb
               program? */
            if (post_load_source_count <= a) {
                elog(LOGDEBUG, "force decorrelate due to lack of reverb "
                     "program sources");
                do_decorrelate = true;
            }

            switch (fprg->type) {
            case FILTERPRG_TYPE_REVERB_TRADITIONAL:
                /* If the reverb program impulse response is in the back half,
                   or the speaker/impulse angle difference is less than 90
                   degrees, apply no early attenuation */
                if (isback(fprg->recv_azimuth[receiver[n]], 180) ||
                    anglediff(fprg->recv_azimuth[receiver[n]], 0,
                              azimuth[n], 0) <= 90)
                {
                    elog(LOGDEBUG, "no azimuth early attenation applied");
                    early_attenuate_length = 0.0;
                }
                if ((elevation[n] < -45 &&
                     fprg->recv_elevation[receiver[n]] <= 40 &&
                     fprg->recv_elevation[receiver[n]] >= -20) ||
                    (elevation[n] > 45 &&
                     fprg->recv_elevation[receiver[n]] <= 20 &&
                     fprg->recv_elevation[receiver[n]] >= -20))
                {
                    elog(LOGDEBUG, "applying elevation correction");
                    early_attenuate_length += elevation_correction;
                }
            
                b = a;
                if (post_load_source_count <= a) {
                    b = post_load_source_count - 1;
                }
                for (i = 0; i < prepf.n_rates; i++) {
                    dest[i] = prepf.filters[b][i];
                    destlen[i] = prepf.lengths[i];
                }
                break;
                
            case FILTERPRG_TYPE_REVERB_BFORMAT:

                /* Full 3D, we do not need early attenuation compensation */
                early_attenuate_length = 0.0;
                
                if (buf[0] == NULL) {
                    for (i = 0; i < prepf.n_rates; i++) {
                        buf[i] = alignalloc(prepf.lengths[i] * realsize);
                        buflengths[i] = prepf.lengths[i];
                        if (buf[i] == NULL) {
                            err = VCU_ERR_NO_MEMORY;
                            goto reverb_preprocess_abort;
                        }
                    }
                } else {
                    for (i = 0; i < prepf.n_rates; i++) {
                        if (prepf.lengths[i] != buflengths[i]) {
                            err = VCU_ERR_INVALID_FORMAT;
                            goto reverb_preprocess_abort;
                        }
                    }
                }

                /* compensate for the receiver axis rotation */
                x = cos((azimuth[n] - fprg->recv_azimuth[receiver[n]]) /
                        180.0 * M_PI) *
                    cos(elevation[n] / 180.0 * M_PI);
                y = sin((azimuth[n] - fprg->recv_azimuth[receiver[n]]) /
                        180.0 * M_PI) *
                    cos(elevation[n] / 180.0 * M_PI);
                z = sin(elevation[n] / 180 * M_PI);
                xf = (float)x;
                yf = (float)y;
                zf = (float)z;

                b = 4 * a;
                if (post_load_source_count <= a) {
                    b = 4 * (post_load_source_count - 1);
                }
                if (fprg->subtype == FILTERPRG_SUBTYPE_BFORMAT_NORMALISED) {
                    elog(LOGDEBUG, "Normalised B-format, no W scaling.");
                    iwmulf = 1.0;
                    iwmul = 1.0;
                } else {
                    elog(LOGDEBUG, "Normal B-format (W -3dB): restoring W.");
                    iwmulf = powf(10, 3.0 / 20.0);
                    iwmul = pow(10, 3.0 / 20.0);
                }
                d = settings.bformat_directivity;
                df = (float)settings.bformat_directivity;
                for (i = 0; i < target[0].n_rates; i++) {
                    if (realsize == 4) {
                        for (k = 0; k < prepf.lengths[i]; k++) {
                            iwf = ((float *)prepf.filters[b+0][i])[k];
                            ixf = ((float *)prepf.filters[b+1][i])[k];
                            iyf = ((float *)prepf.filters[b+2][i])[k];
                            izf = ((float *)prepf.filters[b+3][i])[k];
                            ((float *)buf[i])[k] =
                                0.5 * ((2.0 - df) * iwf * iwmulf +
                                       df * (xf * ixf + yf * iyf + zf * izf));
                        }
                        dest[i] = buf[i];
                    } else {
                        for (k = 0; k < prepf.lengths[i]; k++) {
                            iw = ((double *)prepf.filters[b+0][i])[k];
                            ix = ((double *)prepf.filters[b+1][i])[k];
                            iy = ((double *)prepf.filters[b+2][i])[k];
                            iz = ((double *)prepf.filters[b+3][i])[k];
                            ((double *)buf[i])[k] =
                                0.5 * ((2.0 - d) * iw * iwmulf +
                                       d * (x * ix + y * iy + z * iz));
                        }
                        dest[i] = buf[i];
                    }
                    destlen[i] = prepf.lengths[i];
                }
                break;
            }
            
            elog(LOGDEBUG, "generating impulse index %d",
                 settings.n_sources * n + a);
            elog(LOGDEBUG, "speaker %d, az %g, el %g",
                 n, azimuth[n], elevation[n]);
            elog(LOGDEBUG, "reverb engine: %d sources, current index %d",
                 settings.n_sources, a);
            elog(LOGDEBUG, "reverb program impulse index %d, az %g el %g",
                 receiver[n], fprg->recv_azimuth[receiver[n]],
                 fprg->recv_elevation[receiver[n]]);
            elog(LOGDEBUG, "decorrelate %d, fade_in %g, early_attenuate %g",
                 do_decorrelate, fade_in_length, early_attenuate_length);

            sprintf(dump_filename, "log/t%d", settings.n_sources * n + a);
            if (progress_callback != NULL) {
                progress_callback(settings.n_sources * j + a,
                                  settings.n_sources * n_reverb_speakers,
                                  -1, -1, true);
            }
            if ((err = apply_response(&target[settings.n_sources*n+a],
                                      (const void **)dest,
                                      destlen,
                                      fprg->predelay,
                                      target_length,
                                      do_decorrelate,
                                      settings.force_decay,
                                      fade_in_length,
                                      0.0,
                                      early_attenuate_length,
                                      &tail_volume,
                                      &length,
                                      &rt60,
                                      &_truncated,
                                      prepf.realsize,
                                      NULL)) != 0)
            {
                elog(LOGDEBUG, "apply_response() failed");
                goto reverb_preprocess_abort;
            }
            *new_tail_volume += tail_volume;
            if (length > *new_length) {
                *new_length = length;
            }
            if (_truncated) {
                *truncated = true;
            }
            if ((j == 0 && a == 0) || *new_rt60 > 0) {
                if (rt60 == 0) {
                    *new_rt60 = 0;
                } else {
                    *new_rt60 += rt60;
                }
            }
        }
    }
    *orig_rt60 /= prepf_count;
    *orig_tail_volume /= prepf_count;
    *new_rt60 /= settings.n_sources * n_reverb_speakers;
    *new_tail_volume /= settings.n_sources * n_reverb_speakers;
    elog(LOGDEBUG, "tail volume: %f dB; length: %f s; rt60: %f s",
         *new_tail_volume, *new_length, *new_rt60);

    err = 0;
 reverb_preprocess_abort:
    for (n = 0; n < prepf.n_rates; n++) {
        free(buf[n]);
    }
    filterprg_free_prepfilter_contents(&prepf);
    
    return err;
}
