/*
 * (c) Copyright 2002 - 2005 -- Anders Torger
 *
 * This program is open source. For license terms, see the LICENSE file.
 *
 */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <inttypes.h>
#include <sys/time.h>
#include <time.h>

#define index _index
#define ALSA_PCM_NEW_HW_PARAMS_API
#define ALSA_PCM_NEW_SW_PARAMS_API
#include <alsa/asoundlib.h>
#undef index

#include <bfmod.h>

#include "defs.h"
#include "apui_defs.h"
#include "elog.h"
#include "sysmon.h"

struct cpustat {
    uint64_t user;
    uint64_t nice;
    uint64_t system;
    uint64_t idle;
};

static bool_t
get_cpu_stats(struct cpustat stats[],
              int *n_cpus)
{
    char s[1024], *p;
    FILE *stream;
    int n;
    
    if ((stream = fopen("/proc/stat", "rt")) == NULL) {
        return false;
    }
    s[1023] = '\0';
    if ((fgets(s, 1023, stream)) == NULL) {
        fclose(stream);
        return false;
    }
    s[4] = '\0';
    if (strcmp(s, "cpu ") != 0) {
        fclose(stream);
        return false;
    }
    if ((fgets(s, 1023, stream)) == NULL) {
        fclose(stream);
        return false;
    }
    n = 0;
    while (strstr(s, "cpu") == s) {
        stats[n].user = strtoll(&s[4], &p, 10);
        if (p == NULL) {
            fclose(stream);
            return false;
        }
        stats[n].nice = strtoll(p, &p, 10);
        if (p == NULL) {
            fclose(stream);
            return false;
        }
        stats[n].system = strtoll(p, &p, 10);
        if (p == NULL) {
            fclose(stream);
            return false;
        }
        stats[n].idle = strtoll(p, &p, 10);
        if (p == NULL) {
            fclose(stream);
            return false;
        }
        if ((fgets(s, 1023, stream)) == NULL) {
            fclose(stream);
            return false;
        }
        n++;
    }
    fclose(stream);

    if (n_cpus != NULL) {
        *n_cpus = n;
    }
    return true;
}

static void
cpu_load(const struct cpustat *start,
         const struct cpustat *stop,
         double *usertime,
         double *systemtime)
{
    uint64_t usr, sys, idl, tot;

    usr = stop->user + stop->nice - start->user - start->nice;
    sys = stop->system - start->system;
    idl = stop->idle - start->idle;
    tot = usr + sys + idl;
    *usertime = (double)usr / (double)tot;
    *systemtime = (double)sys / (double)tot;
}

static int
alsa_sample_rate(const char device_params[],
                 int open_channels,
                 int sample_format,
		 int suggested_sample_rate)
{
    int err, format, actual_rate, sample_size, frame_count, n;
    int rates[VCU_VALID_RATES_COUNT] = VCU_VALID_RATES, measured_rate;
    struct timeval tv, starttv, endtv;
    snd_pcm_uframes_t period_size;
    snd_pcm_hw_params_t *params;
    bool_t isinterleaved;
    snd_pcm_t *handle;
    void *buf, **bufs;
    double secs, rate;
    unsigned int un;

    buf = NULL;
    if ((err = snd_pcm_open(&handle, device_params,
			    SND_PCM_STREAM_CAPTURE, 0)) < 0)
    {
	elog(LOGDEBUG, "Could not open audio \"%s\": %s.", device_params,
	     snd_strerror(err));
	/* assume no input signal */
	return SYSMON_NA;
    }
    snd_pcm_hw_params_alloca(&params);
    if ((err = snd_pcm_hw_params_any(handle, params)) < 0) {
	elog(LOGDEBUG, "Could not get any hardware configuration: %s.",
	     snd_strerror(err));
	/* assume no sample clock */
	return 0;
    }
    un = (unsigned int)suggested_sample_rate;
    if ((err = snd_pcm_hw_params_set_rate_near(handle, params,
					       &un, NULL)) < 0)
    {
	elog(LOGDEBUG, "Failed to suggest sample rate %d Hz: %s.\n",
	     suggested_sample_rate, snd_strerror(err));
	/* assume no sample clock */
	return 0;
    }
    actual_rate = un;

    /*
      Since some sound card drivers are a bit funky, we cannot trust the
      reported sample rate. We simply open the input and measure the sample
      clock
     */
    switch (sample_format) {
    case BF_SAMPLE_FORMAT_S8:
	format = SND_PCM_FORMAT_S8;
	sample_size = 1;
	break;
    case BF_SAMPLE_FORMAT_S16_LE:
	format = SND_PCM_FORMAT_S16_LE;
	sample_size = 2;
	break;
    case BF_SAMPLE_FORMAT_S16_BE:
	format = SND_PCM_FORMAT_S16_BE;
	sample_size = 2;
	break;
    case BF_SAMPLE_FORMAT_S24_LE:
	format = SND_PCM_FORMAT_S24_LE;
	sample_size = 3;
	break;
    case BF_SAMPLE_FORMAT_S24_BE:
	format = SND_PCM_FORMAT_S24_BE;
	sample_size = 3;
	break;
    case BF_SAMPLE_FORMAT_S16_4LE:
    case BF_SAMPLE_FORMAT_S24_4LE:
    case BF_SAMPLE_FORMAT_S32_LE:
	format = SND_PCM_FORMAT_S32_LE;
	sample_size = 4;
	break;
    case BF_SAMPLE_FORMAT_S16_4BE:
    case BF_SAMPLE_FORMAT_S24_4BE:
    case BF_SAMPLE_FORMAT_S32_BE:
	format = SND_PCM_FORMAT_S32_BE;
	sample_size = 4;
	break;
    case BF_SAMPLE_FORMAT_FLOAT_LE:
	format = SND_PCM_FORMAT_FLOAT_LE;
	sample_size = 4;
	break;
    case BF_SAMPLE_FORMAT_FLOAT_BE:
	format = SND_PCM_FORMAT_FLOAT_BE;
	sample_size = 4;
	break;
    case BF_SAMPLE_FORMAT_FLOAT64_LE:
	format = SND_PCM_FORMAT_FLOAT64_LE;
	sample_size = 8;
	break;
    case BF_SAMPLE_FORMAT_FLOAT64_BE:
	format = SND_PCM_FORMAT_FLOAT64_BE;
	sample_size = 8;
	break;
    default:
	elog(LOGDEBUG, "Unsupported sample format %d.", sample_format);
        goto return_result;
    }
    if ((err = snd_pcm_hw_params_set_format(handle, params, format)) < 0) {
	elog(LOGDEBUG, "Failed to set sample format to %s: %s.",
             bf_strsampleformat(sample_format), 
             snd_strerror(err));
        goto return_result;
    }
    if (snd_pcm_hw_params_set_access(handle, params,
				     SND_PCM_ACCESS_RW_INTERLEAVED) < 0)
    {
	if ((err =
	     snd_pcm_hw_params_set_access(handle, params,
					  SND_PCM_ACCESS_RW_NONINTERLEAVED))
	    < 0)
	{
	    elog(LOGDEBUG, "Failed to set interleaved and non-interleaved "
		    "access mode: %s.", snd_strerror(err));
            goto return_result;
	}
	isinterleaved = false;
    } else {
	isinterleaved = true;
    }
    if ((err = snd_pcm_hw_params_set_channels(handle, params,
					      open_channels)) < 0)
    {
	elog(LOGDEBUG, "Failed to set channel count to %d: %s.",
             open_channels, snd_strerror(err));
	return false;
    }
    period_size = 64;
    snd_pcm_hw_params_set_period_size_near(handle, params, &period_size, NULL);
    buf = malloc(period_size * open_channels * sample_size);
    bufs = alloca(open_channels * sizeof(void *));
    memset(bufs, 0, open_channels * sizeof(void *));
    bufs[0] = buf;

    if (snd_pcm_hw_params(handle, params) < 0) {
	elog(LOGDEBUG, "Unable to install hw params\n");
        goto return_result;
    }
    if ((err = snd_pcm_prepare(handle)) < 0) {
	elog(LOGDEBUG, "Unable to prepare audio: %s.", snd_strerror(err));
        goto return_result;
    }
    
    if ((err = snd_pcm_start(handle)) < 0) {
        elog(LOGDEBUG, "ALSA I/O: Could not start audio: %s.",
                snd_strerror(err));
        goto return_result;
    }
    if (isinterleaved) {
        err = snd_pcm_readi(handle, buf, period_size);
    } else {    
        err = snd_pcm_readn(handle, bufs, period_size);
    }

    frame_count = 0;
    tv.tv_sec = 1;
    tv.tv_usec = 0;
    gettimeofday(&starttv, NULL);
    timeradd(&starttv, &tv, &endtv);
    do {
        if (isinterleaved) {
            err = snd_pcm_readi(handle, buf, period_size);
        } else {    
            err = snd_pcm_readn(handle, bufs, period_size);
        }
        if (err < 0) {
            elog(LOGDEBUG, "ALSA I/O: Could not read audio: %s.",
                 snd_strerror(err));
            goto return_result;
        }
        frame_count += err;
        gettimeofday(&tv, NULL);
    } while (timercmp(&tv, &endtv, <));
    endtv = tv;
    timersub(&endtv, &starttv, &tv);
    secs = ((double)tv.tv_sec * 1000000.0 + (double)tv.tv_usec) / 1000000.0;

    rate = (double)frame_count / secs;
    elog(LOGDEBUG, "got %d frames in %f seconds (rate: %f)", frame_count, secs,
         rate);
    
    for (n = 0; n < VCU_VALID_RATES_COUNT; n++) {
        if ((double)rates[n] > rate * 0.98 &&
            (double)rates[n] < rate / 0.98)
        {
            measured_rate = rates[n];
            break;
        }
    }
    if (n == VCU_VALID_RATES_COUNT) {
        elog(LOGDEBUG, "measured rate did not match a valid rate");
        goto return_result;
    }
    elog(LOGDEBUG, "driver said %d, measured %d", actual_rate, rates[n]);
    actual_rate = rates[n];

 return_result:
    free(buf);
    if ((err = snd_pcm_close(handle)) < 0) {
	elog(LOGDEBUG, "Failed to close device: %s.", snd_strerror(err));
	/* continue anyway  */
    }
    elog(LOGDEBUG, "sample rate: %d", actual_rate);
    return actual_rate;
}

static const char *
brutefir_alsa_param(const char param[])
{
    static char str[100];
    char *p;
    
    /* FIXME: this is a bit ugly, needing to to brutefir parsing here... */
    if ((p = strchr(param, '\"')) == NULL) {
        return NULL;
    }
    strncpy(str, p + 1, 100);
    str[99] = '\0';
    if ((p = strchr(str, '\"')) == NULL) {
        return NULL;
    }
    *p = '\0';
    return str;
}

int
sysmon_verify_bfir_device(const char bfir_device[],
                          const char bfir_device_params[],
                          bool_t isinput)
{
    snd_pcm_uframes_t min_period_size;
    snd_pcm_hw_params_t *params;
    snd_pcm_t *handle = NULL;
    int err;
    
    if (strcmp(bfir_device, "alsa") != 0) {
        return BFIR_MIN_BLOCKSIZE_LOWLAT;
    }
    if ((err = snd_pcm_open(&handle, brutefir_alsa_param(bfir_device_params),
			    (isinput) ? SND_PCM_STREAM_CAPTURE :
                            SND_PCM_STREAM_PLAYBACK, 0)) < 0)
    {
	elog(LOGERROR, "Could not open audio \"%s\": %s (%d). "
             "Verify static configuration",
             brutefir_alsa_param(bfir_device_params),
	     snd_strerror(err), err);
        exit(EXIT_FAILURE);
    }

    snd_pcm_hw_params_alloca(&params);
    if ((err = snd_pcm_hw_params_any(handle, params)) < 0) {
	elog(LOGERROR, "Could not get any hardware configuration: %s.\n",
             snd_strerror(err));
        exit(EXIT_FAILURE);
    }
    snd_pcm_hw_params_get_period_size_min(params, &min_period_size, NULL);
    
    snd_pcm_close(handle);

    return (int)min_period_size;
}

void
sysmon_sysinfo(struct basic_sysinfo *sysinfo)
{
    FILE *stream;
    char s[1000], *p;
    
    if (sysinfo == NULL) {
	return;
    }
    memset(sysinfo, 0, sizeof(struct basic_sysinfo));

    if ((stream = fopen("/proc/cpuinfo", "rt")) == NULL) {
	return;
    }
    s[999] = '\0';
    while (fgets(s, 999, stream) != NULL) {
	if (strncasecmp(s, "processor", 9) == 0) {
	    sysinfo->cpu_count++;
	}
	if (strncasecmp(s, "model name", 10) == 0 &&
	    (p = strchr(s, ':')) != NULL &&
	    sysinfo->cpu_modelname[0] == '\0')
	{
	    p++;
	    while (*p == ' ' || *p == '\t') p++;
	    strncpy(sysinfo->cpu_modelname, p, sizeof(sysinfo->cpu_modelname));
	    sysinfo->cpu_modelname[sizeof(sysinfo->cpu_modelname)-1] = '\0';
	    if ((p = strchr(sysinfo->cpu_modelname, '\n')) != NULL) {
		*p = '\0';
	    }
	    p = sysinfo->cpu_modelname;
	    while (*p != '\0') {
		if (*p == '\"') {
		    *p = '\'';
		}
		p++;
	    }
	}
	if (strncasecmp(s, "cpu mhz", 7) == 0 &&
	    (p = strchr(s, ':')) != NULL &&
	    sysinfo->cpu_speed == 0)
	{
	    sysinfo->cpu_speed = atoi(++p);
	}
    }
    fclose(stream);
    
    if ((stream = fopen("/proc/meminfo", "rt")) == NULL) {
	return;
    }
    while (fgets(s, 999, stream) != NULL) {
	if (strncasecmp(s, "memtotal", 8) == 0 &&
	    (p = strchr(s, ':')) != NULL)
	{
	    sysinfo->mem_total = atoi(++p);
	    break;
	}
    }
    fclose(stream);

    /*
    elog("cpu_count: %d", sysinfo->cpu_count);
    elog("cpu_speed: %d", sysinfo->cpu_speed);
    elog("cpu_modelname: \"%s\"", sysinfo->cpu_modelname);
    elog("mem_total: %d", sysinfo->mem_total);
    */
}

void
sysmon_uptime(int *days,
	      int *hours)
{
    uint64_t seconds = 0;
    FILE *stream;
    char s[1000];

    *days = 0;
    *hours = 0;
    if ((stream = fopen("/proc/uptime", "rt")) == NULL) {
	return;
    }
    s[999] = '\0';
    if (fgets(s, 999, stream) != NULL) {
	seconds = atoll(s);
    }
    fclose(stream);
    *days = (int)(seconds / 86400);
    *hours = (int)((seconds - (seconds / 86400) * 86400) / 3600);
}

int
sysmon_sampling_rate(const char bfir_device[],
		     const char bfir_device_params[],
                     int open_channels,
                     int sample_format,
		     int suggested_sample_rate)
{
    if (strcmp(bfir_device, "alsa") == 0) {
        elog(LOGDEBUG, "returning alsa sample rate");
	return alsa_sample_rate(brutefir_alsa_param(bfir_device_params),
                                open_channels, sample_format,
                                suggested_sample_rate);
    } else if (strcmp(bfir_device, "file") == 0) {
        elog(LOGDEBUG, "file device -- no sample clock");
	return 0;
    } else {
	elog(LOGDEBUG, "unknown device type \"%s\", assume no sample clock",
             bfir_device);
	return 0;
    }
}

int
sysmon_n_cpus(void)
{
    FILE *stream;
    char s[1000];
    int n_cpus = 0;

    /* if we fail somewhere, we report that there is one processor */
    
    if ((stream = fopen("/proc/cpuinfo", "rt")) == NULL) {
	return 1;
    }
    s[999] = '\0';
    while (fgets(s, 999, stream) != NULL) {
	if (strncasecmp(s, "processor", 9) == 0) {
	    n_cpus++;
	}
    }
    fclose(stream);
    if (n_cpus == 0) {
	n_cpus = 1;
    }
    return n_cpus;
}

double
sysmon_cpuload(void)
{
    static struct cpustat stats[2][64];
    static int cur = 0;
    static bool_t isinit = false;
    double usertime, systime, maxload;
    int n, n_cpus;

    if (!isinit) {
        memset(stats, 0, sizeof(stats));
        get_cpu_stats(stats[1], NULL);
        isinit = true;
        return 0.0;
    }
    maxload = 0.0;
    get_cpu_stats(stats[cur], &n_cpus);
    for (n = 0; n < n_cpus; n++) {
        cpu_load(&stats[!cur][n], &stats[cur][n], &usertime, &systime);
        if (maxload < usertime + systime) {
            maxload = usertime + systime;
        }
    }
    cur = !cur;
    return (double)maxload;
}

double
sysmon_memused(void)
{
    int memtotal, memfree, buffers, cached;
    FILE *stream;
    char s[100];
    
    if ((stream = fopen("/proc/meminfo", "rt")) == NULL) {
	return 0.0;
    }
    s[999] = '\0';
    memtotal = memfree = buffers = cached = 0;
    s[sizeof(s)-1] = '\0';
    while (fgets(s, sizeof(s)-1, stream) != NULL) {
        if (memtotal == 0 && strstr(s, "MemTotal:") == s) {
            memtotal = atoi(&s[9]);
        } else if (memfree == 0 && strstr(s, "MemFree:") == s) {
            memfree = atoi(&s[8]);
        } else if (buffers == 0 && strstr(s, "Buffers:") == s) {
            buffers = atoi(&s[8]);
        } else if (cached == 0 && strstr(s, "Cached:") == s) {
            cached = atoi(&s[7]);
        }
        if (memtotal != 0 && memfree != 0 && buffers != 0 && cached != 0) {
            break;
        }
    }
    fclose(stream);
    return (double)(memtotal - memfree - buffers - cached) / (double)memtotal;
}

double
sysmon_temp(const char path[])
{
    FILE *stream;
    char s[1024], *p;

    if (path == NULL || path[0] == '\0') {
        return SYSMON_NA;
    }
    if ((stream = fopen(path, "rt")) == NULL) {
        elog(LOGDEBUG, "could not open file \"%s\" for reading: %s",
             path, strerror(errno));
        return SYSMON_NA;
    }
    s[1023] = '\0';
    if (fgets(s, 1023, stream) == NULL) {
        fclose(stream);
        return SYSMON_NA;
    }
    if ((p = strrchr(s, ' ')) == NULL) {
        /* assume there is no status bit */
        return atof(s);
    }
    if (atoi(s) == 0) {
        return SYSMON_NA;
    }
    return atof(p);
}

#define DUMPCMD(str) \
    sprintf(cmd, "echo " str " >> %s 2>&1", path); system(cmd); \
    sprintf(cmd, str " >> %s 2>&1", path); system(cmd);

bool_t
sysmon_dumpinfo(const char path[])
{
    char cmd[PATH_MAX + 200];

    if (strlen(path) > PATH_MAX) {
        return false;
    }

    sprintf(cmd, "cat /proc/version > %s 2>&1", path);
    system(cmd);
    DUMPCMD("date");
    DUMPCMD("cat /proc/cpuinfo");
    DUMPCMD("cat /proc/meminfo");
    DUMPCMD("cat /proc/devices");
    DUMPCMD("cat /proc/dma");
    DUMPCMD("cat /proc/filesystems");
    DUMPCMD("cat /proc/interrupts");
    DUMPCMD("cat /proc/iomem");
    DUMPCMD("cat /proc/ioports");
    DUMPCMD("cat /proc/loadavg");
    DUMPCMD("cat /proc/misc");
    DUMPCMD("cat /proc/modules");
    DUMPCMD("cat /proc/mounts");
    DUMPCMD("cat /proc/partitions");
    DUMPCMD("cat /proc/pci");
    DUMPCMD("lspci");
    DUMPCMD("cat /proc/slabinfo");
    DUMPCMD("cat /proc/stat");
    DUMPCMD("cat /proc/swaps");
    DUMPCMD("cat /proc/uptime");
    DUMPCMD("cat /proc/asound/cards");
    /* FIXME: implement proper traversing of asound tree */
    DUMPCMD("cat /proc/asound/card0/rme9652");
    DUMPCMD("ps auxwww");
    DUMPCMD("df");

    return true;
}

#undef DUMPCMD
