/*
 * (c) Copyright 2002 - 2004 -- Anders Torger
 *
 * This program is open source. For license terms, see the LICENSE file.
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <curses.h>
#include <unistd.h>
#include <string.h>

#include "defs.h"
#include "twki.h"
#include "textgui.h"
#include "almusvcu.h"
#include "elog.h"

static int xsize, ysize;
static WINDOW *win = NULL;
static int keytab[KI_KEY_MAX + 1];

static void
draw_border(void)
{
    int n;
    
    mvwaddch(win, 0, 0, ACS_ULCORNER);
    mvwaddch(win, ysize + 1, 0, ACS_LLCORNER);
    mvwaddch(win, 0, xsize + 1, ACS_URCORNER);
    mvwaddch(win, ysize + 1, xsize + 1, ACS_LRCORNER);
    for (n = 1; n <= xsize; n++) {
	mvwaddch(win, 0, n, ACS_HLINE);
	mvwaddch(win, ysize + 1, n, ACS_HLINE);
    }
    for (n = 1; n <= ysize; n++) {
	mvwaddch(win, n, 0, ACS_VLINE);
	mvwaddch(win, n, xsize + 1, ACS_VLINE);
    }
}

static void
twin_properties(int xsize_range[2],
		int ysize_range[2],
		char direction_chars[4])
{
    xsize_range[0] = 2;
    xsize_range[1] = 78;
    ysize_range[0] = 2;
    ysize_range[1] = 50;
    direction_chars[0] = '<';
    direction_chars[1] = '>';
    direction_chars[2] = '^';
    direction_chars[3] = 'v';
}

static bool_t
twin_init(int _xsize,
	  int _ysize)
{
    xsize = _xsize;
    ysize = _ysize;

    initscr();
    cbreak();
    nonl();
    curs_set(0);
    win = newwin(ysize + 2, xsize + 2, 0, 0);
    intrflush(win, false);
    draw_border();
    return true;
}

static void
twin_clear(void)
{
    wclear(win);
    draw_border();
}

static void
twin_putchar(int x,
	     int y,
	     char c)
{    
    mvwaddch(win, y + 1, x + 1, c);
}

static void
twin_putstring(int x,
	       int y,
	       const char s[])
{
    mvwaddstr(win, y + 1, x + 1, s);
}

static void
twin_refresh(void)
{
    wrefresh(win);
}

#define PARSEKEY(keyname)                                                      \
    if (strcasecmp(s, #keyname) == 0) {                                        \
        return keyname;                                                        \
    }

#define _PARSEKEY(keyname, keydefine)                                          \
    if (strcasecmp(s, keyname) == 0) {                                         \
        return keydefine;                                                      \
    }

static int
parsekey(const char s[])
{
    char strkeyname[10];
    int n;
    
    if (strlen(s) == 1) {
        return s[0];
    }
    PARSEKEY(KEY_LEFT);
    PARSEKEY(KEY_RIGHT);
    PARSEKEY(KEY_UP);
    PARSEKEY(KEY_DOWN);
    PARSEKEY(KEY_HOME);
    PARSEKEY(KEY_BACKSPACE);
    for (n = 0; n < 64; n++) {
        sprintf(strkeyname, "KEY_F%d", n);
        _PARSEKEY(strkeyname, KEY_F(n));
    }
    PARSEKEY(KEY_DL);
    PARSEKEY(KEY_IL);
    PARSEKEY(KEY_DC);
    PARSEKEY(KEY_IC);
    PARSEKEY(KEY_EIC);
    PARSEKEY(KEY_CLEAR);
    PARSEKEY(KEY_EOS);
    PARSEKEY(KEY_EOL);
    PARSEKEY(KEY_SF);
    PARSEKEY(KEY_SR);
    PARSEKEY(KEY_NPAGE);
    PARSEKEY(KEY_PPAGE);
    PARSEKEY(KEY_STAB);
    PARSEKEY(KEY_CTAB);
    PARSEKEY(KEY_CATAB);
    PARSEKEY(KEY_ENTER);
    PARSEKEY(KEY_PRINT);
    PARSEKEY(KEY_LL);
    PARSEKEY(KEY_A1);
    PARSEKEY(KEY_A3);
    PARSEKEY(KEY_B2);
    PARSEKEY(KEY_C1);
    PARSEKEY(KEY_C3);
    PARSEKEY(KEY_BTAB);
    PARSEKEY(KEY_BEG);
    PARSEKEY(KEY_CANCEL);
    PARSEKEY(KEY_CLOSE);
    PARSEKEY(KEY_COMMAND);
    PARSEKEY(KEY_COPY);
    PARSEKEY(KEY_CREATE);
    PARSEKEY(KEY_END);
    PARSEKEY(KEY_EXIT);
    PARSEKEY(KEY_FIND);
    PARSEKEY(KEY_HELP);
    PARSEKEY(KEY_MARK);
    PARSEKEY(KEY_MESSAGE);
    PARSEKEY(KEY_MOVE);
    PARSEKEY(KEY_NEXT);
    PARSEKEY(KEY_OPEN);
    PARSEKEY(KEY_OPTIONS);
    PARSEKEY(KEY_PREVIOUS);
    PARSEKEY(KEY_REDO);
    PARSEKEY(KEY_REFERENCE);
    PARSEKEY(KEY_REFRESH);
    PARSEKEY(KEY_REPLACE);
    PARSEKEY(KEY_RESTART);
    PARSEKEY(KEY_RESUME);
    PARSEKEY(KEY_SAVE);
    PARSEKEY(KEY_SBEG);
    PARSEKEY(KEY_SCANCEL);
    PARSEKEY(KEY_SCOMMAND);
    PARSEKEY(KEY_SCOPY);
    PARSEKEY(KEY_SCREATE);
    PARSEKEY(KEY_SDC);
    PARSEKEY(KEY_SDL);
    PARSEKEY(KEY_SELECT);
    PARSEKEY(KEY_SEND);
    PARSEKEY(KEY_SEOL);
    PARSEKEY(KEY_SEXIT);
    PARSEKEY(KEY_SFIND);
    PARSEKEY(KEY_SHELP);
    PARSEKEY(KEY_SHOME);
    PARSEKEY(KEY_SIC);
    PARSEKEY(KEY_SLEFT);
    PARSEKEY(KEY_SMESSAGE);
    PARSEKEY(KEY_SMOVE);
    PARSEKEY(KEY_SNEXT);
    PARSEKEY(KEY_SOPTIONS);
    PARSEKEY(KEY_SPREVIOUS);
    PARSEKEY(KEY_SPRINT);
    PARSEKEY(KEY_SREDO);
    PARSEKEY(KEY_SREPLACE);
    PARSEKEY(KEY_SRIGHT);
    PARSEKEY(KEY_SRSUME);
    PARSEKEY(KEY_SSAVE);
    PARSEKEY(KEY_SSUSPEND);
    PARSEKEY(KEY_SUNDO);
    PARSEKEY(KEY_SUSPEND);
    PARSEKEY(KEY_UNDO);
    return -1;
}

static int
keyin_init(void)
{
    keypad(win, true);
    noecho();
    
    return STDIN_FILENO;
}

static int
keyin_getkey(int fd,
	     int *repeats)
{
    int key, n;

    *repeats = 0;
    key = wgetch(win);

    if (key == ERR) {
        elog(LOGERROR, "wgetch failed");
        return KI_KEY_FAILURE;
    }
    
    for (n = 0; n <= KI_KEY_MAX; n++) {
        if (key == keytab[n]) {
            return n;
        }
    }
    return KI_KEY_IGNORE;
}

static struct twki twki_curses = {
    {
        twin_properties,
        twin_putchar,
        twin_init,
        twin_refresh,
        twin_clear,
        twin_putstring,
        NULL
    },
    {
        keyin_init,
        keyin_getkey
    }
};

void
vcu_finish(void)
{
    elog(LOGDEBUG, "vcu_finish() called");
    endwin();
}

const char *
vcu_name(void)
{
    return "twki_curses";
}

bool_t
vcu_needterm(void)
{
    return true;
}

void *
vcu_share(void)
{
    return (void *)&twki_curses;
}

#define KEYFIELD(keyname, keydefine)                                           \
    if (strcmp(lval.field, keyname) == 0) {                                    \
        if (get_config_token(&lval) != VCU_LEXVAL_STRING) {                    \
            elog(LOGERROR, "Parse error: expected string");                    \
            return false;                                                      \
        }                                                                      \
        if ((n = parsekey(lval.string)) == -1) {                               \
            elog(LOGERROR, "Parse error: unknown key \"%s\"", lval.string);    \
            return false;                                                      \
        }                                                                      \
        keytab[keydefine] = n;                                                 \
        if (get_config_token(&lval) != VCU_LEX_EOS) {                          \
            elog(LOGERROR, "Parse error: expected end of statement (;)");      \
            return false;                                                      \
        }                                                                      \
        continue;                                                              \
    }

bool_t
vcu_preinit(int (*get_config_token)(union vcu_lexval *lexval))
{
    union vcu_lexval lval;
    int tok, n, i;

    memset(keytab, -1, sizeof(keytab));
    keytab[KI_KEY_UP] = KEY_UP;
    keytab[KI_KEY_DOWN] = KEY_DOWN;
    keytab[KI_KEY_LEFT] = KEY_LEFT;
    keytab[KI_KEY_RIGHT] = KEY_RIGHT;
    while ((tok = get_config_token(&lval)) > 0) {
        if (tok != VCU_LEXVAL_FIELD) {
	    elog(LOGERROR, "Parse error: expected field.");
	    return false;
        }

        KEYFIELD("key_left", KI_KEY_LEFT);
        KEYFIELD("key_right", KI_KEY_RIGHT);
        KEYFIELD("key_up", KI_KEY_UP);
        KEYFIELD("key_down", KI_KEY_DOWN);
        KEYFIELD("key_mvol_up", KI_KEY_MVOL_UP);
        KEYFIELD("key_mvol_down", KI_KEY_MVOL_DOWN);
        KEYFIELD("key_gvol_up", KI_KEY_GVOL_UP);
        KEYFIELD("key_gvol_down", KI_KEY_GVOL_DOWN);
        KEYFIELD("key_reverb_prg_next", KI_KEY_REVERB_PRG_NEXT);
        KEYFIELD("key_channel_mode_next", KI_KEY_CHANNEL_MODE_NEXT);
        KEYFIELD("key_group_next", KI_KEY_GROUP_NEXT);
        KEYFIELD("key_eq_mode_next", KI_KEY_EQ_MODE_NEXT);
        KEYFIELD("key_master_mute", KI_KEY_MASTER_MUTE);
        KEYFIELD("key_group_mute", KI_KEY_GROUP_MUTE);

        elog(LOGERROR, "Parse error: unknown field \"%s\"", lval.field);
        return false;
    }

    for (n = 0; n <= KI_KEY_MAX; n++) {
        for (i = 0; i <= KI_KEY_MAX; i++) {
            if (n != i && keytab[n] != -1 && keytab[n] == keytab[i]) {
                elog(LOGERROR, "At least one key is used more than once");
                return false;
            }
        }
    }
    return true;
}

bool_t
vcu_init(const struct vcu *_vcu)
{
    int textgui_index = -1, self_index = -1;
    const struct ui_module *modules;
    int n_modules, n;

    modules = _vcu->ctl.get_modules(&n_modules);
    for (n = 0; n < n_modules; n++) {
	if (strcmp(modules[n].name, "textgui") == 0) {
	    textgui_index = n;
	}
	if (strcmp(modules[n].name, "twki_curses") == 0) {
	    self_index = n;
	}
    }
    if (self_index == -1) {
	elog(LOGERROR, "Did not find module \"twki_curses\".");
	return false;
    }
    if (textgui_index == -1) {
	elog(LOGERROR, "Module \"twki_curses\" depends on module \"textgui\".");
	return false;
    }
    if (textgui_index < self_index) {
	elog(LOGERROR, "Module \"twki_curses\" must be loaded "
             "before module \"textgui\".");
	return false;
    }
    return true;
}
